"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016 Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging
from datetime import datetime

from ofxparse import AccountType, OfxParser
from pytz import UTC

from biweeklybudget.db import db_session, upsert_record
from biweeklybudget.models.ofx_transaction import OFXTransaction
from biweeklybudget.models.ofx_statement import OFXStatement
from biweeklybudget.models.account import Account
from biweeklybudget.utils import dtnow

logger = logging.getLogger(__name__)


class DuplicateFileException(Exception):
    """
    Exception raised when trying to parse a file that has already been parsed
    for the Account (going by the OFX signon date).
    """
    pass


class OFXUpdater(object):
    """
    Class to wrap updating the database with a parsed OFX file.
    """

    def __init__(self, acct_id, acct_name, cat_memo=False):
        """
        Initialize OFXUpdater for a specified account.

        :param acct_id: account database ID
        :type acct_id: int
        :param acct_name: account name
        :type acct_name: str
        :param cat_memo: whether or not to concatenate OFX Memo to Name
        :type cat_memo: bool
        """
        logger.debug(
            'Initializing OFXUpdater for account "%s" (%d)', acct_name, acct_id
        )
        self.acct_id = acct_id
        self.acct_name = acct_name
        self.acct = db_session.query(Account).get(self.acct_id)
        self.cat_memo = cat_memo

    def update(self, ofx, mtime=None, filename=None):
        """
        Update a single OFX file for this account.

        :param ofx: Ofx instance for parsed file
        :type ofx: ``ofxparse.ofxparse.Ofx``
        :param mtime: OFX file modification time (or current time)
        :type mtime: datetime.datetime
        :param filename: OFX file name
        :type filename: str
        :returns: the OFXStatement created by this run
        :rtype: biweeklybudget.models.ofx_statement.OFXStatement
        """
        if mtime is None:
            mtime = dtnow()
        if hasattr(ofx, 'status') and ofx.status['severity'] == 'ERROR':
            raise RuntimeError("OFX Error: %s" % vars(ofx))
        stmt = self._create_statement(ofx, mtime, filename)
        if ofx.account.type == AccountType.Bank:
            stmt.type = 'Bank'
            return self._update_bank_or_credit(ofx, stmt)
        elif ofx.account.type == AccountType.CreditCard:
            stmt.type = 'CreditCard'
            return self._update_bank_or_credit(ofx, stmt)
        elif ofx.account.type == AccountType.Investment:
            return self._update_investment(ofx, stmt)
        raise RuntimeError("Don't know how to update AccountType %d",
                           ofx.account.type)

    def _create_statement(self, ofx, mtime, filename):
        """
        Create an OFXStatement for this OFX file. If one already exists with
        the same account and filename, raise DuplicateFileException.

        :param ofx: Ofx instance for parsed file
        :type ofx: ``ofxparse.ofxparse.Ofx``
        :param mtime: OFX file modification time (or current time)
        :type mtime: datetime.datetime
        :param filename: OFX file name
        :type filename: str
        :return: the OFXStatement object
        :rtype: biweeklybudget.models.ofx_statement.OFXStatement
        :raises: DuplicateFileException
        """
        ofx_date = OfxParser.parseOfxDateTime(
            ofx.signon.dtserver).replace(tzinfo=UTC)
        stmt = db_session.query(OFXStatement).filter(
            OFXStatement.account_id == self.acct.id,
            OFXStatement.filename == filename
        ).first()
        if stmt is not None:
            logger.debug('Found existing statement with same as_of date, '
                         'id=%d; raising DuplicateFileException()', stmt.id)
            raise DuplicateFileException()
        logger.debug('Creating new OFXStatement account_id=%s filename=%s '
                     'as_of=%s', self.acct.id, filename, ofx_date)
        a = OFXStatement(
            account_id=self.acct.id,
            filename=filename,
            file_mtime=mtime,
            as_of=ofx_date,
            currency=ofx.account.curdef,
            acctid=ofx.account.account_id
        )
        if ofx.account.institution is not None:
            a.bankid = ofx.account.institution.fid
        if ofx.account.routing_number is not None:
            a.routing_number = ofx.account.routing_number
        if ofx.account.account_type is not None:
            a.acct_type = ofx.account.account_type
        if hasattr(ofx.account, 'brokerid'):
            a.brokerid = ofx.account.brokerid
        return a

    def _update_bank_or_credit(self, ofx, stmt):
        """
        Update a single OFX file for this Bank or Credit account.

        :param ofx: Ofx instance for parsed file
        :type ofx: ``ofxparse.ofxparse.Ofx``
        :param stmt: the OFXStatement for this statement
        :type stmt: biweeklybudget.models.ofx_statement.OFXStatement
        :returns: the OFXStatement object
        :rtype: biweeklybudget.models.ofx_statement.OFXStatement
        """
        # Note that as of 0.16, OfxParser returns tz-naive UTC datetimes
        logger.debug('Updating Bank/Credit account')
        if hasattr(ofx.account.statement, 'available_balance'):
            stmt.avail_bal = ofx.account.statement.available_balance
        if hasattr(ofx.account.statement, 'available_balance_date'):
            stmt.avail_bal_as_of = \
                ofx.account.statement.available_balance_date.replace(tzinfo=UTC)
        stmt.ledger_bal = ofx.account.statement.balance
        stmt.ledger_bal_as_of = \
            ofx.account.statement.balance_date.replace(tzinfo=UTC)
        db_session.add(stmt)
        self.acct.set_balance(
            overall_date=stmt.as_of,
            ledger=stmt.ledger_bal,
            ledger_date=stmt.ledger_bal_as_of,
            avail=stmt.avail_bal,
            avail_date=stmt.avail_bal_as_of
        )
        for txn in ofx.account.statement.transactions:
            try:
                kwargs = OFXTransaction.params_from_ofxparser_transaction(
                    txn, self.acct.id, stmt, cat_memo=self.cat_memo
                )
            except RuntimeError as ex:
                logger.error(ex)
                continue
            upsert_record(
                OFXTransaction,
                ['account_id', 'fitid'],
                **kwargs
            )
        return stmt

    def _update_investment(self, ofx, stmt):
        """
        Update a single OFX file for this Investment account.

        :param ofx: Ofx instance for parsed file
        :type ofx: ``ofxparse.ofxparse.Ofx``
        :param stmt: the OFXStatement for this statement
        :type stmt: biweeklybudget.models.ofx_statement.OFXStatement
        :returns: the OFXStatement object
        :rtype: biweeklybudget.models.ofx_statement.OFXStatement
        """
        logger.debug('Updating Investment account')
        stmt.type = 'Investment'
        value = 0
        earliest_dt = datetime.max
        for pos in ofx.account.statement.positions:
            if hasattr(pos, 'date') and isinstance(pos.date, datetime):
                if pos.date < earliest_dt:
                    earliest_dt = pos.date
            if hasattr(pos, 'market_value'):
                value += pos.market_value
            else:
                value += (pos.units * pos.unit_price)
        if earliest_dt != datetime.max:
            stmt.ledger_bal_as_of = earliest_dt.replace(tzinfo=UTC)
        if value != 0:
            stmt.ledger_bal = value
        db_session.add(stmt)
        self.acct.set_balance(
            overall_date=stmt.as_of,
            ledger=stmt.ledger_bal,
            ledger_date=stmt.ledger_bal_as_of
        )
        return stmt



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016 Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import logging
from flask.views import MethodView
from flask import render_template, jsonify

from biweeklybudget.flaskapp.app import app
from biweeklybudget.db import db_session
from biweeklybudget.models.budget_model import Budget
from biweeklybudget.flaskapp.views.formhandlerview import FormHandlerView

logger = logging.getLogger(__name__)


class BudgetsView(MethodView):
    """
    Render the GET /budgets view using the ``budgets.html`` template.
    """

    def get(self):
        standing = db_session.query(Budget).filter(
            Budget.is_periodic.__eq__(False)
        ).order_by(Budget.name).all()
        periodic = db_session.query(Budget).filter(
            Budget.is_periodic.__eq__(True)
        ).order_by(Budget.name).all()
        return render_template(
            'budgets.html',
            standing=standing,
            periodic=periodic
        )


class OneBudgetView(MethodView):
    """
    Render the GET /budgets/<int:budget_id> view using the ``budgets.html``
    template.
    """

    def get(self, budget_id):
        standing = db_session.query(Budget).filter(
            Budget.is_periodic.__eq__(False)
        ).order_by(Budget.name).all()
        periodic = db_session.query(Budget).filter(
            Budget.is_periodic.__eq__(True)
        ).order_by(Budget.name).all()
        return render_template(
            'budgets.html',
            standing=standing,
            periodic=periodic,
            budget_id=budget_id
        )


class BudgetAjax(MethodView):
    """
    Handle GET /ajax/budget/<int:budget_id> endpoint.
    """

    def get(self, budget_id):
        budget = db_session.query(Budget).get(budget_id)
        return jsonify(budget.as_dict)


class BudgetFormHandler(FormHandlerView):
    """
    Handle POST /forms/budget
    """

    def validate(self, data):
        """
        Validate the form data. Return None if it is valid, or else a hash of
        field names to list of error strings for each field.

        :param data: submitted form data
        :type data: dict
        :return: None if no errors, or hash of field name to errors for that
          field
        """
        have_errors = False
        errors = {k: [] for k in data.keys()}
        if data.get('name', '').strip() == '':
            errors['name'].append('Name cannot be empty')
            have_errors = True
        if (
            data['is_periodic'] == 'true' and
            data['starting_balance'].strip() == ''
        ):
            errors['starting_balances'].append(
                'Starting balance must be specified for periodic budgets.'
            )
            have_errors = True
        if (
            data['is_periodic'] == 'false' and
            data['current_balance'].strip() == ''
        ):
            errors['current_balances'].append(
                'Current balance must be specified for standing budgets.'
            )
            have_errors = True
        if have_errors:
            return errors
        return None

    def submit(self, data):
        """
        Handle form submission; create or update models in the DB. Raises an
        Exception for any errors.

        :param data: submitted form data
        :type data: dict
        :return: message describing changes to DB (i.e. link to created record)
        :rtype: str
        """
        if 'id' in data and data['id'].strip() != '':
            # updating an existing budget
            budget = db_session.query(Budget).get(int(data['id']))
            if budget is None:
                raise RuntimeError("Error: no Budget with ID %s" % data['id'])
            action = 'updating Budget ' + data['id']
        else:
            budget = Budget()
            action = 'creating new Budget'
        budget.name = data['name'].strip()
        budget.description = data['description'].strip()
        if data['is_periodic'] == 'true':
            budget.is_periodic = True
            budget.starting_balance = data['starting_balance']
        else:
            budget.is_periodic = False
            budget.current_balance = data['current_balance']
        if data['is_active'] == 'true':
            budget.is_active = True
        else:
            budget.is_active = False
        if data['is_income'] == 'true':
            budget.is_income = True
        else:
            budget.is_income = False
        logger.info('%s: %s', action, budget.as_dict)
        db_session.add(budget)
        db_session.commit()
        return 'Successfully saved Budget %d in database.' % budget.id


app.add_url_rule('/budgets', view_func=BudgetsView.as_view('budgets_view'))
app.add_url_rule(
    '/budgets/<int:budget_id>',
    view_func=OneBudgetView.as_view('one_budget_view')
)
app.add_url_rule(
    '/ajax/budget/<int:budget_id>',
    view_func=BudgetAjax.as_view('budget_ajax')
)
app.add_url_rule(
    '/forms/budget',
    view_func=BudgetFormHandler.as_view('budget_form')
)



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016 Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

from datetime import datetime
import os
import argparse
import logging
import atexit
from copy import deepcopy
from io import StringIO
import importlib

from ofxparse import OfxParser
from ofxclient.account import Account as OfxClientAccount

from biweeklybudget.utils import Vault
from biweeklybudget import settings
from biweeklybudget.models.account import Account
from biweeklybudget.db import init_db, db_session, cleanup_db
from biweeklybudget.ofxupdater import OFXUpdater
from biweeklybudget.cliutils import set_log_debug, set_log_info

logger = logging.getLogger(__name__)

# suppress requests logging
requests_log = logging.getLogger("requests")
requests_log.setLevel(logging.WARNING)
requests_log.propagate = True


class OfxGetter(object):

    @staticmethod
    def accounts():
        """
        Return a sorted list of all Account objects that are for ofxgetter.
        """
        return [a for a in db_session.query(
                Account).filter(
            Account.for_ofxgetter).order_by(Account.name).all()
        ]

    def __init__(self, savedir='./'):
        self.savedir = savedir
        self._account_data = {}
        for acct in self.accounts():
            if acct.vault_creds_path is None and acct.ofxgetter_config == {}:
                continue
            self._account_data[acct.name] = {
                'vault_path': acct.vault_creds_path,
                'config': acct.ofxgetter_config,
                'id': acct.id,
                'cat_memo': acct.ofx_cat_memo_to_name
            }
        logger.info('Initialized with data for %d accounts',
                    len(self._account_data))
        self._accounts = {}
        self.vault = Vault()
        for acct_name in self._account_data.keys():
            data = self._account_data[acct_name]['config']
            vault_path = self._account_data[acct_name]['vault_path']
            logger.debug('Getting secrets for account %s', acct_name)
            secrets = self.vault.read(vault_path)
            data['institution']['password'] = secrets['password']
            data['institution']['username'] = secrets['username']
            if 'class_name' not in data:
                self._accounts[acct_name] = OfxClientAccount.deserialize(data)
        logger.info('Initialized %d accounts', len(self._accounts))
        self.now_str = datetime.now().strftime('%Y-%m-%d_%H-%M-%S')

    def get_ofx(self, account_name, write_to_file=True, days=30):
        """
        Download OFX from the specified account. Return it as a string.

        :param account_name: account name to download
        :type account_name: str
        :param write_to_file: if True, also write to a file named
          "<account_name>_<date stamp>.ofx"
        :type write_to_file: bool
        :param days: number of days of data to download
        :type days: int
        :return: OFX string
        :rtype: str
        """
        fname = None
        logger.info('Downloading OFX for account: %s', account_name)
        if 'class_name' in self._account_data[account_name]['config']:
            ofxdata = self._get_ofx_scraper(account_name, days=days)
        else:
            acct = self._accounts[account_name]
            logger.info(
                'Disabling logging for ofxclient, which has bad logging'
            )
            oldlvl = logging.getLogger().getEffectiveLevel()
            logging.getLogger().setLevel(logging.WARNING)
            ofxdata = acct.download(days=days).read()
            logging.getLogger().setLevel(oldlvl)
            logger.info('Re-enabling logging')
        if write_to_file:
            fname = self._write_ofx_file(account_name, ofxdata)
        self._ofx_to_db(account_name, fname, ofxdata)
        return ofxdata

    def _ofx_to_db(self, account_name, fname, ofxdata):
        """
        Put OFX Data to the DB

        :param account_name: account name to download
        :type account_name: str
        :param ofxdata: raw OFX data
        :type ofxdata: str
        :param fname: filename OFX was written to
        :type fname: str
        """
        logger.debug('Parsing OFX')
        ofx = OfxParser.parse(StringIO(ofxdata))
        logger.debug('Instantiating OFXUpdater')
        updater = OFXUpdater(
            self._account_data[account_name]['id'],
            account_name,
            cat_memo=self._account_data[account_name]['cat_memo']
        )
        logger.debug('Updating OFX in DB')
        updater.update(ofx, filename=fname)
        db_session.commit()
        logger.debug('Done updating OFX in DB')

    def _get_ofx_scraper(self, account_name, days=30):
        """
        Get OFX via a ScreenScraper subclass.

        :param account_name: account name
        :type account_name: str
        :param days: number of days of data to download
        :type days: int
        :return: OFX string
        :rtype: str
        """
        data = self._account_data[account_name]['config']
        clsname = data['class_name']
        modname = data['module_name']
        logger.debug('Scraper - getting class %s from module %s',
                     clsname, modname)
        cls = getattr(
            importlib.import_module(modname),
            clsname
        )
        logger.debug('Getting secrets for account %s', account_name)
        secrets = self.vault.read(
            self._account_data[account_name]['vault_path']
        )
        if 'kwargs' in data:
            kwargs = deepcopy(data['kwargs'])
        else:
            kwargs = {}
        kwargs['username'] = secrets['username']
        kwargs['password'] = secrets['password']
        kwargs['savedir'] = os.path.join(self.savedir, account_name)
        acct = cls(**kwargs)
        ofxdata = acct.run()
        return ofxdata

    def _write_ofx_file(self, account_name, ofxdata):
        """
        Write OFX data to a file.

        :param account_name: account name
        :type account_name: str
        :param ofxdata: raw OFX data string
        :type ofxdata: str
        :returns: name of the file that was written
        :rtype: str
        """
        if not os.path.exists(os.path.join(self.savedir, account_name)):
            os.makedirs(os.path.join(self.savedir, account_name))
        fname = '%s_%s.ofx' % (account_name, self.now_str)
        fpath = os.path.join(self.savedir, account_name, fname)
        logger.debug('Writing %d bytes of OFX to: %s', len(ofxdata), fpath)
        with open(fpath, 'w') as fh:
            fh.write(ofxdata)
        logger.info('Wrote OFX data to: %s', fpath)
        return fname


def parse_args():
    p = argparse.ArgumentParser(description='Download OFX transactions')
    p.add_argument('-v', '--verbose', dest='verbose', action='count', default=0,
                   help='verbose output. specify twice for debug-level output.')
    p.add_argument('-l', '--list-accts', dest='list', action='store_true',
                   help='list accounts and exit')
    p.add_argument('ACCOUNT_NAME', type=str, action='store', default=None,
                   nargs='?',
                   help='Account name; omit to download all accounts')
    args = p.parse_args()
    return args


def main():
    global logger
    format = "[%(asctime)s %(levelname)s] %(message)s"
    logging.basicConfig(level=logging.WARNING, format=format)
    logger = logging.getLogger()

    args = parse_args()

    # set logging level
    if args.verbose > 1:
        set_log_debug(logger)
    elif args.verbose == 1:
        set_log_info(logger)

    atexit.register(cleanup_db)
    init_db()

    if args.list:
        for k in OfxGetter.accounts():
            print(k.name)
        raise SystemExit(0)

    getter = OfxGetter(settings.STATEMENTS_SAVE_PATH)
    if args.ACCOUNT_NAME is not None:
        getter.get_ofx(args.ACCOUNT_NAME)
        raise SystemExit(0)
    # else all of them
    total = 0
    success = 0
    for acct in OfxGetter.accounts():
        try:
            total += 1
            getter.get_ofx(acct.name)
            success += 1
        except Exception:
            logger.error(
                'Failed to download account %s', acct.name, exc_info=True
            )
    if success != total:
        logger.warning('Downloaded %d of %d accounts', success, total)
        raise SystemExit(1)
    raise SystemExit(0)


if __name__ == "__main__":
    main()



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016 Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me ia email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import pytest
import os
from datetime import timedelta, datetime
from dateutil.relativedelta import relativedelta
from pytz import UTC
from selenium.webdriver.support.ui import Select

from biweeklybudget.utils import dtnow
from biweeklybudget.tests.acceptance_helpers import AcceptanceHelper
from biweeklybudget.settings import PAY_PERIOD_START_DATE
from biweeklybudget.biweeklypayperiod import BiweeklyPayPeriod
from biweeklybudget.models import *
import biweeklybudget.models.base  # noqa
from biweeklybudget.tests.conftest import engine

dt = dtnow()


@pytest.mark.acceptance
class TestPayPeriods(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium, testflask, refreshdb):  # noqa
        self.baseurl = base_url
        self.get(selenium, base_url + '/payperiods')

    def test_heading(self, selenium):
        heading = selenium.find_element_by_class_name('navbar-brand')
        assert heading.text == 'Pay Periods - BiweeklyBudget'

    def test_nav_menu(self, selenium):
        ul = selenium.find_element_by_id('side-menu')
        assert ul is not None
        assert 'nav' in ul.get_attribute('class')
        assert ul.tag_name == 'ul'

    def test_notifications(self, selenium):
        div = selenium.find_element_by_id('notifications-row')
        assert div is not None
        assert div.get_attribute('class') == 'row'


@pytest.mark.acceptance
class TestPayPeriodFor(AcceptanceHelper):

    def test_current_period(self, base_url, selenium):
        start_date = PAY_PERIOD_START_DATE
        print("PayPeriod start date: %s" % start_date)
        self.get(
            selenium,
            base_url + '/pay_period_for?date=' + start_date.strftime('%Y-%m-%d')
        )
        self.wait_for_load_complete(selenium)
        assert selenium.current_url == \
            base_url + '/payperiod/' + start_date.strftime('%Y-%m-%d')

    def test_current_period_end(self, base_url, selenium):
        start_date = PAY_PERIOD_START_DATE
        print("PayPeriod start date: %s" % start_date)
        send_date = start_date + timedelta(days=10)
        self.get(
            selenium,
            base_url + '/pay_period_for?date=' + send_date.strftime('%Y-%m-%d')
        )
        self.wait_for_load_complete(selenium)
        assert selenium.current_url == \
            base_url + '/payperiod/' + start_date.strftime('%Y-%m-%d')


@pytest.mark.acceptance
class TestFindPayPeriod(AcceptanceHelper):

    def test_input_date(self, base_url, selenium):
        self.get(selenium, base_url + '/payperiods')
        i = selenium.find_element_by_id('payperiod_date_input')
        i.clear()
        start_date = PAY_PERIOD_START_DATE
        print("PayPeriod start date: %s" % start_date)
        send_date = start_date + timedelta(days=4)
        i.send_keys(send_date.strftime('%Y-%m-%d'))
        selenium.find_element_by_id('payperiod-go-button').click()
        self.wait_for_load_complete(selenium)
        assert selenium.current_url == \
            base_url + '/payperiod/' + start_date.strftime('%Y-%m-%d')

    def test_input_calendar(self, base_url, selenium):
        self.get(selenium, base_url + '/payperiods')
        start_date = PAY_PERIOD_START_DATE
        print("PayPeriod start date: %s" % start_date)
        send_date = start_date + timedelta(days=4)
        if send_date.month > dtnow().month:
            print("Using datepicker cal3")
            div_id = 'cal3'
            exp_month_th = send_date.strftime('%B %Y')
        else:
            print("Using datepicker cal2")
            div_id = 'cal2'
            exp_month_th = dtnow().strftime('%B %Y')
        daysdiv = selenium.find_element_by_xpath(
            '//div[@id="%s"]//div[@class="datepicker-days"]' % div_id
        )
        tbl = daysdiv.find_elements_by_tag_name('table')[0]
        # month
        assert tbl.find_elements_by_tag_name(
            'thead')[0].find_elements_by_tag_name(
            'tr')[1].find_elements_by_tag_name('th')[1].text == exp_month_th
        tbody = tbl.find_elements_by_tag_name('tbody')[0]
        print('Looking for datepicker TD for date %s' % send_date)
        for e in tbody.find_elements_by_tag_name('td'):
            if (
                e.get_attribute('class') != 'day' and
                e.get_attribute('class') != 'today day'
            ):
                continue
            if e.text.strip() == str(send_date.day):
                parent = e.find_element_by_xpath('..')
                print("Found date TD: %s - parent: %s" % (
                    e.get_attribute('innerHTML'),
                    parent.get_attribute('innerHTML')
                ))
                e.click()
                break
        else:
            raise RuntimeError("Unable to find td for date %d", send_date.day)
        self.wait_for_load_complete(selenium)
        assert selenium.current_url == \
            base_url + '/payperiod/' + start_date.strftime('%Y-%m-%d')

    def test_input_calendar_future(self, base_url, selenium):
        self.get(selenium, base_url + '/payperiods')
        start_date = dtnow()
        print("PayPeriod start date: %s" % start_date)
        send_date = start_date + relativedelta(months=3)
        send_date = send_date.replace(day=3)
        print("send_date: %s" % send_date)
        send_pp = BiweeklyPayPeriod.period_for_date(send_date, None)
        print("send_pp.start_date: %s" % send_pp.start_date)
        daysdiv = selenium.find_element_by_xpath(
            '//div[@id="cal2"]//div[@class="datepicker-days"]'
        )
        tbl = daysdiv.find_elements_by_tag_name('table')[0]
        thead = tbl.find_elements_by_tag_name('thead')[0]
        # month
        assert thead.find_elements_by_tag_name(
            'tr')[1].find_elements_by_tag_name('th')[1].text == \
            dtnow().strftime('%B %Y')
        tbody = tbl.find_elements_by_tag_name('tbody')[0]
        assets_dir = os.path.abspath(os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            '..', '..', '..', '..', '..',
            'results', 'assets'
        ))
        ss_path = os.path.join(assets_dir, 'test_input_calendar_future1.png')
        print('Screenshotting to: %s' % ss_path)
        selenium.get_screenshot_as_file(ss_path)
        print('Advancing by 3 months')
        next_link = thead.find_element_by_xpath('//th[@class="next"]')
        next_link.click()
        self.wait_for_jquery_done(selenium)
        next_link.click()
        self.wait_for_jquery_done(selenium)
        next_link.click()
        self.wait_for_jquery_done(selenium)
        ss_path = os.path.join(assets_dir, 'test_input_calendar_future2.png')
        print('Screenshotting to: %s' % ss_path)
        selenium.get_screenshot_as_file(ss_path)
        print('Looking for datepicker TD for date %s' % send_date)
        for e in tbody.find_elements_by_tag_name('td'):
            if e.get_attribute('class') != 'day':
                continue
            if e.text.strip() == str(send_date.day):
                parent = e.find_element_by_xpath('..')
                print("Found date TD: %s - parent: %s" % (
                    e.get_attribute('innerHTML'),
                    parent.get_attribute('innerHTML')
                ))
                e.click()
                break
        else:
            raise RuntimeError("Unable to find td for date %d", send_date.day)
        self.wait_for_load_complete(selenium)
        assert selenium.current_url == \
            base_url + '/payperiod/' + send_pp.start_date.strftime('%Y-%m-%d')


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb')
class TestPayPeriodsIndex(AcceptanceHelper):

    def test_0_clean_db(self, testdb):
        # clean the database
        biweeklybudget.models.base.Base.metadata.reflect(engine)
        biweeklybudget.models.base.Base.metadata.drop_all(engine)
        biweeklybudget.models.base.Base.metadata.create_all(engine)

    def test_1_add_account(self, testdb):
        a = Account(
            description='First Bank Account',
            name='BankOne',
            ofx_cat_memo_to_name=True,
            ofxgetter_config_json='{"foo": "bar"}',
            vault_creds_path='secret/foo/bar/BankOne',
            acct_type=AcctType.Bank
        )
        testdb.add(a)
        a.set_balance(
            overall_date=datetime(2017, 4, 10, 12, 0, 0, tzinfo=UTC),
            ledger=1.0,
            ledger_date=datetime(2017, 4, 10, 12, 0, 0, tzinfo=UTC)
        )
        testdb.flush()
        testdb.commit()

    def test_2_add_budgets(self, testdb):
        testdb.add(Budget(
            name='1Income',
            is_periodic=True,
            description='1Income',
            starting_balance=1000.00,
            is_income=True
        ))
        testdb.add(Budget(
            name='2Periodic',
            is_periodic=True,
            description='2Periodic',
            starting_balance=500.00
        ))
        testdb.add(Budget(
            name='3Periodic',
            is_periodic=True,
            description='3Periodic',
            starting_balance=0.00
        ))
        testdb.flush()
        testdb.commit()

    def pay_periods(self, db):
        return {
            'prev': BiweeklyPayPeriod.period_for_date(
                (PAY_PERIOD_START_DATE - timedelta(days=2)), db),
            'curr': BiweeklyPayPeriod.period_for_date(
                PAY_PERIOD_START_DATE, db),
            'next': BiweeklyPayPeriod.period_for_date(
                (PAY_PERIOD_START_DATE + timedelta(days=15)), db),
            'following': BiweeklyPayPeriod.period_for_date(
                (PAY_PERIOD_START_DATE + timedelta(days=29)), db)
        }

    def test_3_add_transactions(self, testdb):
        acct = testdb.query(Account).get(1)
        ibudget = testdb.query(Budget).get(1)
        e1budget = testdb.query(Budget).get(2)
        e2budget = testdb.query(Budget).get(3)
        periods = self.pay_periods(testdb)
        # previous pay period
        ppdate = periods['prev'].start_date
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=1)),
            actual_amount=100.00,
            budgeted_amount=100.00,
            description='prev income',
            account=acct,
            budget=ibudget
        ))
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=2)),
            actual_amount=250.00,
            description='prev trans 1',
            account=acct,
            budget=e2budget
        ))
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=3)),
            actual_amount=600.00,
            budgeted_amount=500.00,
            description='prev trans 2',
            account=acct,
            budget=e1budget
        ))
        ppdate = periods['curr'].start_date
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=1)),
            actual_amount=1400.00,
            budgeted_amount=100.00,
            description='prev income',
            account=acct,
            budget=ibudget
        ))
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=2)),
            actual_amount=1850.00,
            description='prev trans 1',
            account=acct,
            budget=e2budget
        ))
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=3)),
            actual_amount=600.00,
            budgeted_amount=500.00,
            description='prev trans 2',
            account=acct,
            budget=e1budget
        ))
        ppdate = periods['next'].start_date
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=1)),
            actual_amount=1400.00,
            budgeted_amount=100.00,
            description='prev income',
            account=acct,
            budget=ibudget
        ))
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=2)),
            actual_amount=788.00,
            description='prev trans 1',
            account=acct,
            budget=e2budget
        ))
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=3)),
            actual_amount=600.00,
            budgeted_amount=500.00,
            description='prev trans 2',
            account=acct,
            budget=e1budget
        ))
        ppdate = periods['following'].start_date
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=1)),
            actual_amount=1400.00,
            budgeted_amount=100.00,
            description='prev income',
            account=acct,
            budget=ibudget
        ))
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=2)),
            actual_amount=2.00,
            description='prev trans 1',
            account=acct,
            budget=e2budget
        ))
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=3)),
            actual_amount=600.00,
            budgeted_amount=500.00,
            description='prev trans 2',
            account=acct,
            budget=e1budget
        ))
        testdb.flush()
        testdb.commit()

    def test_4_confirm_sums(self, testdb):
        periods = self.pay_periods(testdb)
        assert periods['prev'].overall_sums == {
            'allocated': 750.0,
            'spent': 850.0,
            'income': 1000.0,
            'remaining': 150.0
        }
        assert periods['curr'].overall_sums == {
            'allocated': 2350.0,
            'spent': 2450.0,
            'income': 1400.0,
            'remaining': -1050.0
        }
        assert periods['next'].overall_sums == {
            'allocated': 1288.0,
            'spent': 1388.0,
            'income': 1400.0,
            'remaining': 12.0
        }
        assert periods['following'].overall_sums == {
            'allocated': 502.0,
            'spent': 602.0,
            'income': 1400.0,
            'remaining': 798.0
        }

    def test_5_pay_periods_table(self, base_url, selenium, testdb):
        periods = self.pay_periods(testdb)
        self.get(selenium, base_url + '/payperiods')
        table = selenium.find_element_by_id('pay-period-table')
        texts = self.tbody2textlist(table)
        elems = self.tbody2elemlist(table)
        assert texts == [
            [
                periods['prev'].start_date.strftime('%Y-%m-%d'),
                '$750.00',
                '$850.00',
                '$150.00'
            ],
            [
                periods['curr'].start_date.strftime('%Y-%m-%d') + ' (current)',
                '$2,350.00',
                '$2,450.00',
                '-$1,050.00'
            ],
            [
                periods['next'].start_date.strftime('%Y-%m-%d'),
                '$1,288.00',
                '$1,388.00',
                '$12.00'
            ],
            [
                periods['following'].start_date.strftime('%Y-%m-%d'),
                '$502.00',
                '$602.00',
                '$798.00'
            ]
        ]
        # test links
        links = [x[0].get_attribute('innerHTML') for x in elems]
        expected = []
        for k in ['prev', 'curr', 'next', 'following']:
            dstr = periods[k].start_date.strftime('%Y-%m-%d')
            s = '<a href="/payperiod/%s">%s</a>' % (dstr, dstr)
            if k == 'curr':
                s += ' <em>(current)</em>'
            expected.append(s)
        assert links == expected
        # test red text for negative dollar amounts
        assert elems[0][3].get_attribute('innerHTML') == '$150.00'
        assert elems[1][3].get_attribute('innerHTML') == '<span ' \
            'class="text-danger">-$1,050.00</span>'
        # test highlighted row for current period
        tbody = table.find_element_by_tag_name('tbody')
        trs = tbody.find_elements_by_tag_name('tr')
        assert trs[1].get_attribute('class') == 'info'

    def test_6_notification_panels(self, base_url, selenium, testdb):
        periods = self.pay_periods(testdb)
        self.get(selenium, base_url + '/payperiods')
        this_panel = selenium.find_element_by_id('panel-period-current')
        assert this_panel.get_attribute('class') == 'panel panel-red'
        assert this_panel.find_element_by_class_name('panel-heading').text\
            == '-$1,050.00\nRemaining this period'
        assert this_panel.find_element_by_tag_name('a').get_attribute(
            'href') == base_url + '/payperiod/' + periods[
            'curr'].start_date.strftime('%Y-%m-%d')
        assert this_panel.find_element_by_class_name('panel-footer')\
            .text == 'View ' + periods['curr'].start_date.strftime(
            '%Y-%m-%d') + ' Pay Period'
        next_panel = selenium.find_element_by_id('panel-period-next')
        assert next_panel.get_attribute('class') == 'panel panel-yellow'
        assert next_panel.find_element_by_class_name('panel-heading').text \
            == '$12.00\nRemaining next period'
        assert next_panel.find_element_by_tag_name('a').get_attribute(
            'href') == base_url + '/payperiod/' + periods[
            'next'].start_date.strftime('%Y-%m-%d')
        assert next_panel.find_element_by_class_name('panel-footer')\
            .text == 'View ' + periods['next'].start_date.strftime(
            '%Y-%m-%d') + ' Pay Period'
        following_panel = selenium.find_element_by_id('panel-period-following')
        assert following_panel.get_attribute('class') == 'panel panel-green'
        assert following_panel.find_element_by_class_name('panel-heading')\
            .text == '$798.00\nRemaining following period'
        assert following_panel.find_element_by_tag_name('a').get_attribute(
            'href') == base_url + '/payperiod/' + periods[
            'following'].start_date.strftime('%Y-%m-%d')
        assert following_panel.find_element_by_class_name('panel-footer')\
            .text == 'View ' + periods['following'].start_date.strftime(
            '%Y-%m-%d') + ' Pay Period'


@pytest.mark.acceptance
class TestPayPeriod(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium, testflask, refreshdb):  # noqa
        self.baseurl = base_url
        self.get(
            selenium,
            base_url + '/payperiod/' +
            PAY_PERIOD_START_DATE.strftime('%Y-%m-%d')
        )

    def test_heading(self, selenium, testdb):
        heading = selenium.find_element_by_class_name('navbar-brand')
        pp = BiweeklyPayPeriod(PAY_PERIOD_START_DATE, testdb)
        assert heading.text == '%s to %s Pay Period - BiweeklyBudget' % (
            pp.start_date.strftime('%Y-%m-%d'),
            pp.end_date.strftime('%Y-%m-%d')
        )

    def test_nav_menu(self, selenium):
        ul = selenium.find_element_by_id('side-menu')
        assert ul is not None
        assert 'nav' in ul.get_attribute('class')
        assert ul.tag_name == 'ul'

    def test_notifications(self, selenium):
        div = selenium.find_element_by_id('notifications-row')
        assert div is not None
        assert div.get_attribute('class') == 'row'


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb')
class TestPayPeriodOtherPeriodInfo(AcceptanceHelper):

    def test_0_clean_db(self, testdb):
        # clean the database
        biweeklybudget.models.base.Base.metadata.reflect(engine)
        biweeklybudget.models.base.Base.metadata.drop_all(engine)
        biweeklybudget.models.base.Base.metadata.create_all(engine)

    def test_1_add_account(self, testdb):
        a = Account(
            description='First Bank Account',
            name='BankOne',
            ofx_cat_memo_to_name=True,
            ofxgetter_config_json='{"foo": "bar"}',
            vault_creds_path='secret/foo/bar/BankOne',
            acct_type=AcctType.Bank
        )
        testdb.add(a)
        a.set_balance(
            overall_date=datetime(2017, 4, 10, 12, 0, 0, tzinfo=UTC),
            ledger=1.0,
            ledger_date=datetime(2017, 4, 10, 12, 0, 0, tzinfo=UTC)
        )
        testdb.flush()
        testdb.commit()

    def test_2_add_budgets(self, testdb):
        testdb.add(Budget(
            name='1Income',
            is_periodic=True,
            description='1Income',
            starting_balance=1000.00,
            is_income=True
        ))
        testdb.add(Budget(
            name='2Periodic',
            is_periodic=True,
            description='2Periodic',
            starting_balance=500.00
        ))
        testdb.add(Budget(
            name='3Periodic',
            is_periodic=True,
            description='3Periodic',
            starting_balance=0.00
        ))
        testdb.flush()
        testdb.commit()

    def pay_periods(self, db):
        return {
            'prev': BiweeklyPayPeriod.period_for_date(
                (PAY_PERIOD_START_DATE - timedelta(days=2)), db),
            'curr': BiweeklyPayPeriod.period_for_date(
                PAY_PERIOD_START_DATE, db),
            'next': BiweeklyPayPeriod.period_for_date(
                (PAY_PERIOD_START_DATE + timedelta(days=15)), db),
            'following': BiweeklyPayPeriod.period_for_date(
                (PAY_PERIOD_START_DATE + timedelta(days=29)), db),
            'last': BiweeklyPayPeriod.period_for_date(
                (PAY_PERIOD_START_DATE + timedelta(days=43)), db)
        }

    def test_3_add_transactions(self, testdb):
        acct = testdb.query(Account).get(1)
        ibudget = testdb.query(Budget).get(1)
        e1budget = testdb.query(Budget).get(2)
        e2budget = testdb.query(Budget).get(3)
        periods = self.pay_periods(testdb)
        # previous pay period
        ppdate = periods['prev'].start_date
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=1)),
            actual_amount=100.00,
            budgeted_amount=100.00,
            description='prev income',
            account=acct,
            budget=ibudget
        ))
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=2)),
            actual_amount=250.00,
            description='prev trans 1',
            account=acct,
            budget=e2budget
        ))
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=3)),
            actual_amount=600.00,
            budgeted_amount=500.00,
            description='prev trans 2',
            account=acct,
            budget=e1budget
        ))
        ppdate = periods['curr'].start_date
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=1)),
            actual_amount=1400.00,
            budgeted_amount=100.00,
            description='curr income',
            account=acct,
            budget=ibudget
        ))
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=2)),
            actual_amount=1850.00,
            description='curr trans 1',
            account=acct,
            budget=e2budget
        ))
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=3)),
            actual_amount=600.00,
            budgeted_amount=500.00,
            description='curr trans 2',
            account=acct,
            budget=e1budget
        ))
        ppdate = periods['next'].start_date
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=1)),
            actual_amount=1400.00,
            budgeted_amount=100.00,
            description='next income',
            account=acct,
            budget=ibudget
        ))
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=2)),
            actual_amount=788.00,
            description='next trans 1',
            account=acct,
            budget=e2budget
        ))
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=3)),
            actual_amount=600.00,
            budgeted_amount=500.00,
            description='next trans 2',
            account=acct,
            budget=e1budget
        ))
        ppdate = periods['following'].start_date
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=1)),
            actual_amount=1400.00,
            budgeted_amount=100.00,
            description='following income',
            account=acct,
            budget=ibudget
        ))
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=2)),
            actual_amount=2.00,
            description='following trans 1',
            account=acct,
            budget=e2budget
        ))
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=3)),
            actual_amount=600.00,
            budgeted_amount=500.00,
            description='following trans 2',
            account=acct,
            budget=e1budget
        ))
        ppdate = periods['last'].start_date
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=1)),
            actual_amount=1400.00,
            budgeted_amount=100.00,
            description='last income',
            account=acct,
            budget=ibudget
        ))
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=2)),
            actual_amount=2.00,
            description='last trans 1',
            account=acct,
            budget=e2budget
        ))
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=3)),
            actual_amount=550.00,
            budgeted_amount=550.00,
            description='last trans 2',
            account=acct,
            budget=e1budget
        ))
        testdb.flush()
        testdb.commit()

    def test_4_confirm_sums(self, testdb):
        periods = self.pay_periods(testdb)
        assert periods['prev'].overall_sums == {
            'allocated': 750.0,
            'spent': 850.0,
            'income': 1000.0,
            'remaining': 150.0
        }
        assert periods['curr'].overall_sums == {
            'allocated': 2350.0,
            'spent': 2450.0,
            'income': 1400.0,
            'remaining': -1050.0
        }
        assert periods['next'].overall_sums == {
            'allocated': 1288.0,
            'spent': 1388.0,
            'income': 1400.0,
            'remaining': 12.0
        }
        assert periods['following'].overall_sums == {
            'allocated': 502.0,
            'spent': 602.0,
            'income': 1400.0,
            'remaining': 798.0
        }
        assert periods['last'].overall_sums == {
            'allocated': 552.0,
            'spent': 552.0,
            'income': 1400.0,
            'remaining': 848.0
        }

    def test_5_other_periods_table(self, base_url, selenium, testdb):
        self.get(
            selenium,
            base_url + '/payperiod/' +
            PAY_PERIOD_START_DATE.strftime('%Y-%m-%d')
        )
        pp = BiweeklyPayPeriod(PAY_PERIOD_START_DATE, testdb)
        table = selenium.find_element_by_id('pay-period-table')
        assert self.thead2list(table) == [
            '%s (prev.)' % pp.previous.start_date.strftime('%Y-%m-%d'),
            '%s (curr.)' % pp.start_date.strftime('%Y-%m-%d'),
            '%s (next)' % pp.next.start_date.strftime('%Y-%m-%d'),
            '%s' % pp.next.next.start_date.strftime('%Y-%m-%d'),
            '%s' % pp.next.next.next.start_date.strftime('%Y-%m-%d')
        ]
        assert self.tbody2textlist(table) == [[
            '$150.00',
            '-$1,050.00',
            '$12.00',
            '$798.00',
            '$848.00'
        ]]
        contents = [
            x.get_attribute('innerHTML') for x in self.thead2elemlist(table)
        ]
        assert contents == [
            '<a href="/payperiod/{d}">{d} <em>(prev.)</em></a>'.format(
                d=pp.previous.start_date.strftime('%Y-%m-%d')),
            '{d} <em>(curr.)</em>'.format(
                d=pp.start_date.strftime('%Y-%m-%d')),
            '<a href="/payperiod/{d}">{d} <em>(next)</em></a>'.format(
                d=pp.next.start_date.strftime('%Y-%m-%d')),
            '<a href="/payperiod/{d}">{d}</a>'.format(
                d=pp.next.next.start_date.strftime('%Y-%m-%d')),
            '<a href="/payperiod/{d}">{d}</a>'.format(
                d=pp.next.next.next.start_date.strftime('%Y-%m-%d'))
        ]


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb')
class TestCurrentPayPeriod(AcceptanceHelper):

    def test_00_inactivate_scheduled(self, testdb):
        for s in testdb.query(
                ScheduledTransaction).filter(
            ScheduledTransaction.is_active.__eq__(True)
        ).all():
            s.is_active = False
            testdb.add(s)
        testdb.flush()
        testdb.commit()
        # delete existing transactions
        for tr in testdb.query(TxnReconcile).all():
            testdb.delete(tr)
        for idx in [1, 2, 3]:
            t = testdb.query(Transaction).get(idx)
            testdb.delete(t)
        testdb.flush()
        testdb.commit()

    def test_01_add_transactions(self, testdb):
        acct = testdb.query(Account).get(1)
        e1budget = testdb.query(Budget).get(1)
        e2budget = testdb.query(Budget).get(2)
        pp = BiweeklyPayPeriod.period_for_date(
            PAY_PERIOD_START_DATE, testdb
        )
        ppdate = pp.start_date
        st = ScheduledTransaction(
            account=acct,
            budget=e1budget,
            amount=11.11,
            num_per_period=2,
            description='ST7 per_period'
        )
        testdb.add(st)
        testdb.add(ScheduledTransaction(
            account=acct,
            budget=e1budget,
            amount=22.22,
            day_of_month=(ppdate + timedelta(days=5)).day,
            description='ST8 day_of_month'
        ))
        testdb.add(ScheduledTransaction(
            account=acct,
            budget=e2budget,
            amount=33.33,
            date=(ppdate + timedelta(days=6)),
            description='ST9 date'
        ))
        t = Transaction(
            date=(ppdate + timedelta(days=8)),
            actual_amount=12.00,
            budgeted_amount=11.11,
            description='Txn From ST7',
            account=acct,
            budget=e1budget,
            scheduled_trans=st
        )
        testdb.add(t)
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=6)),
            actual_amount=111.13,
            budgeted_amount=111.11,
            description='T1foo',
            notes='notesT1',
            account=acct,
            scheduled_trans=testdb.query(ScheduledTransaction).get(1),
            budget=e1budget
        ))
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=2)),
            actual_amount=-333.33,
            budgeted_amount=-333.33,
            description='T2',
            notes='notesT2',
            account=testdb.query(Account).get(2),
            scheduled_trans=testdb.query(ScheduledTransaction).get(3),
            budget=testdb.query(Budget).get(4)
        ))
        testdb.add(Transaction(
            date=ppdate,
            actual_amount=222.22,
            description='T3',
            notes='notesT3',
            account=testdb.query(Account).get(3),
            budget=e2budget
        ))
        testdb.flush()
        testdb.commit()
        testdb.add(TxnReconcile(note='foo', txn_id=t.id))
        testdb.flush()
        testdb.commit()

    def test_03_info_panels(self, base_url, selenium, testdb):
        self.get(
            selenium,
            base_url + '/payperiod/' +
            PAY_PERIOD_START_DATE.strftime('%Y-%m-%d')
        )
        assert selenium.find_element_by_id(
            'amt-income').text == '$2,345.67'
        assert selenium.find_element_by_id('amt-allocated').text == '$411.10'
        assert selenium.find_element_by_id('amt-spent').text == '$345.35'
        assert selenium.find_element_by_id('amt-remaining').text == '$1,934.57'

    def test_04_periodic_budgets(self, base_url, selenium, testdb):
        self.get(
            selenium,
            base_url + '/payperiod/' +
            PAY_PERIOD_START_DATE.strftime('%Y-%m-%d')
        )
        table = selenium.find_element_by_id('pb-table')
        elems = self.tbody2elemlist(table)
        htmls = []
        for row in elems:
            htmls.append(
                [x.get_attribute('innerHTML') for x in row]
            )
        assert htmls == [
            [
                '<a href="/budgets/1">Periodic1</a>',
                '$100.00',
                '$155.55',
                '$123.13',
                '<span class="text-danger">-$56.46</span>'
            ],
            [
                '<a href="/budgets/2">Periodic2</a>',
                '$234.00',
                '$255.55',
                '$222.22',
                '<span class="text-danger">-$21.55</span>'
            ],
            [
                '<a href="/budgets/7">Income (i)</a>',
                '$2,345.67',
                '$0.00',
                '$0.00',
                '$2,345.67'
            ]
        ]

    def test_05_standing_budgets(self, base_url, selenium, testdb):
        self.get(
            selenium,
            base_url + '/payperiod/' +
            PAY_PERIOD_START_DATE.strftime('%Y-%m-%d')
        )
        table = selenium.find_element_by_id('sb-table')
        elems = self.tbody2elemlist(table)
        htmls = []
        for row in elems:
            htmls.append(
                [x.get_attribute('innerHTML') for x in row]
            )
        assert htmls == [
            [
                '<a href="/budgets/4">Standing1</a>',
                '$1,284.23'
            ],
            [
                '<a href="/budgets/5">Standing2</a>',
                '$9,482.29'
            ]
        ]

    def test_06_add_trans_button(self, base_url, selenium, testdb):
        self.get(
            selenium,
            base_url + '/payperiod/' +
            PAY_PERIOD_START_DATE.strftime('%Y-%m-%d')
        )
        btn = selenium.find_element_by_id('btn-add-txn')
        btn.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Add New Transaction'

    def test_07_transaction_table(self, base_url, selenium, testdb):
        pp = BiweeklyPayPeriod(PAY_PERIOD_START_DATE, testdb)
        ppdate = pp.start_date
        self.get(
            selenium,
            base_url + '/payperiod/' +
            PAY_PERIOD_START_DATE.strftime('%Y-%m-%d')
        )
        table = selenium.find_element_by_id('trans-table')
        elems = self.tbody2elemlist(table)
        htmls = []
        for row in elems:
            htmls.append(
                [x.get_attribute('innerHTML') for x in row]
            )
        assert htmls == [
            [
                '',
                '$11.11',
                '<em>(sched)</em> <a href="javascript:schedModal(7, null);">'
                'ST7 per_period (7)</a>',
                '<a href="/accounts/1">BankOne</a>',
                '<a href="/budgets/1">Periodic1</a>',
                '<a href="javascript:schedToTransModal(7, \'%s\');">make '
                'trans.</a>' % pp.start_date.strftime('%Y-%m-%d'),
                '&nbsp;'
            ],
            [
                ppdate.strftime('%Y-%m-%d'),
                '$222.22',
                '<a href="javascript:transModal(7, null);">T3 (7)</a>',
                '<a href="/accounts/3">CreditOne</a>',
                '<a href="/budgets/2">Periodic2</a>',
                '&nbsp;',
                '&nbsp;'
            ],
            [
                (ppdate + timedelta(days=2)).strftime('%Y-%m-%d'),
                '-$333.33',
                '<a href="javascript:transModal(6, null);">T2 (6)</a>',
                '<a href="/accounts/2">BankTwoStale</a>',
                '<a href="/budgets/4">Standing1</a>',
                '<em>(from <a href="javascript:schedModal(3, null);">3</a>)'
                '</em>',
                '&nbsp;'
            ],
            [
                (pp.start_date + timedelta(days=5)).strftime('%Y-%m-%d'),
                '$22.22',
                '<em>(sched)</em> <a href="javascript:schedModal(8, null);">'
                'ST8 day_of_month (8)</a>',
                '<a href="/accounts/1">BankOne</a>',
                '<a href="/budgets/1">Periodic1</a>',
                '<a href="javascript:schedToTransModal(8, \'%s\');">make '
                'trans.</a>' % pp.start_date.strftime('%Y-%m-%d'),
                '&nbsp;'
            ],
            [
                (ppdate + timedelta(days=6)).strftime('%Y-%m-%d'),
                '$111.13',
                '<a href="javascript:transModal(5, null);">T1foo (5)</a>',
                '<a href="/accounts/1">BankOne</a>',
                '<a href="/budgets/1">Periodic1</a>',
                '<em>(from <a href="javascript:schedModal(1, null);">1</a>)'
                '</em>',
                '&nbsp;'
            ],
            [
                (pp.start_date + timedelta(days=6)).strftime('%Y-%m-%d'),
                '$33.33',
                '<em>(sched)</em> <a href="javascript:schedModal(9, null);">'
                'ST9 date (9)</a>',
                '<a href="/accounts/1">BankOne</a>',
                '<a href="/budgets/2">Periodic2</a>',
                '<a href="javascript:schedToTransModal(9, \'%s\');">make '
                'trans.</a>' % pp.start_date.strftime('%Y-%m-%d'),
                '&nbsp;'
            ],
            [
                (pp.start_date + timedelta(days=8)).strftime('%Y-%m-%d'),
                '$12.00',
                '<a href="javascript:transModal(4, null);">Txn From ST7'
                ' (4)</a>',
                '<a href="/accounts/1">BankOne</a>',
                '<a href="/budgets/1">Periodic1</a>',
                '<em>(from <a href="javascript:schedModal(7, null);">7</a>)'
                '</em>',
                '<a href="javascript:txnReconcileModal(2)">Yes (2)</a>'
            ]
        ]

    def test_08_transaction_modal(self, base_url, selenium, testdb):
        pp = BiweeklyPayPeriod(PAY_PERIOD_START_DATE, testdb)
        self.get(
            selenium,
            base_url + '/payperiod/' +
            PAY_PERIOD_START_DATE.strftime('%Y-%m-%d')
        )
        link = selenium.find_element_by_xpath('//a[text()="T1foo (5)"]')
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Transaction 5'
        assert body.find_element_by_id(
            'trans_frm_id').get_attribute('value') == '5'
        assert body.find_element_by_id(
            'trans_frm_date').get_attribute('value') == (
            pp.start_date + timedelta(days=6)).strftime('%Y-%m-%d')
        assert body.find_element_by_id(
            'trans_frm_amount').get_attribute('value') == '111.13'
        assert body.find_element_by_id(
            'trans_frm_description').get_attribute('value') == 'T1foo'
        acct_sel = Select(body.find_element_by_id('trans_frm_account'))
        opts = []
        for o in acct_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['1', 'BankOne'],
            ['2', 'BankTwoStale'],
            ['3', 'CreditOne'],
            ['4', 'CreditTwo'],
            ['6', 'DisabledBank'],
            ['5', 'InvestmentOne']
        ]
        assert acct_sel.first_selected_option.get_attribute('value') == '1'
        budget_sel = Select(body.find_element_by_id('trans_frm_budget'))
        opts = []
        for o in budget_sel.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['1', 'Periodic1'],
            ['2', 'Periodic2'],
            ['3', 'Periodic3 Inactive'],
            ['4', 'Standing1'],
            ['5', 'Standing2'],
            ['6', 'Standing3 Inactive'],
            ['7', 'Income (i)']
        ]
        assert budget_sel.first_selected_option.get_attribute('value') == '1'
        assert selenium.find_element_by_id(
            'trans_frm_notes').get_attribute('value') == 'notesT1'

    def test_09_transaction_modal_edit(self, base_url, selenium, testdb):
        self.get(
            selenium,
            base_url + '/payperiod/' +
            PAY_PERIOD_START_DATE.strftime('%Y-%m-%d')
        )
        link = selenium.find_element_by_xpath('//a[text()="T1foo (5)"]')
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Transaction 5'
        assert body.find_element_by_id(
            'trans_frm_id').get_attribute('value') == '5'
        desc = body.find_element_by_id('trans_frm_description')
        desc.send_keys('edited')
        # submit the form
        selenium.find_element_by_id('modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements_by_tag_name('div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Transaction 5 ' \
                                 'in database.'
        # dismiss the modal
        selenium.find_element_by_id('modalCloseButton').click()
        self.wait_for_load_complete(selenium)
        # test that updated budget was removed from the page
        table = selenium.find_element_by_id('trans-table')
        texts = [y[2] for y in self.tbody2textlist(table)]
        assert 'T1fooedited (5)' in texts

    def test_10_transaction_modal_verify_db(self, testdb):
        pp = BiweeklyPayPeriod(PAY_PERIOD_START_DATE, testdb)
        t = testdb.query(Transaction).get(5)
        assert t is not None
        assert t.description == 'T1fooedited'
        assert t.date == (pp.start_date + timedelta(days=6))
        assert float(t.actual_amount) == 111.13
        assert float(t.budgeted_amount) == 111.11
        assert t.account_id == 1
        assert t.budget_id == 1
        assert t.scheduled_trans_id == 1
        assert t.notes == 'notesT1'

    def test_11_sched_trans_verify_db(self, testdb):
        t = testdb.query(ScheduledTransaction).get(7)
        assert t is not None
        assert t.description == 'ST7 per_period'
        assert t.num_per_period == 2
        assert t.date is None
        assert t.day_of_month is None
        assert float(t.amount) == 11.11
        assert t.account_id == 1
        assert t.budget_id == 1
        assert t.notes is None
        assert t.is_active is True

    def test_12_sched_trans_modal(self, base_url, selenium):
        self.get(
            selenium,
            base_url + '/payperiod/' +
            PAY_PERIOD_START_DATE.strftime('%Y-%m-%d')
        )
        link = selenium.find_element_by_xpath(
            '//a[text()="ST7 per_period (7)"]')
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Scheduled Transaction 7'
        assert body.find_element_by_id(
            'sched_frm_id').get_attribute('value') == '7'
        assert body.find_element_by_id(
            'sched_frm_description').get_attribute('value') == 'ST7 per_period'
        assert body.find_element_by_id(
            'sched_frm_type_monthly').is_selected() is False
        assert body.find_element_by_id(
            'sched_frm_type_date').is_selected() is False
        assert body.find_element_by_id(
            'sched_frm_type_per_period').is_selected()
        assert body.find_element_by_id(
            'sched_frm_num_per_period').get_attribute('value') == '2'
        assert body.find_element_by_id(
            'sched_frm_amount').get_attribute('value') == '11.11'
        acct_sel = Select(body.find_element_by_id('sched_frm_account'))
        assert acct_sel.first_selected_option.get_attribute('value') == '1'
        budget_sel = Select(body.find_element_by_id('sched_frm_budget'))
        assert budget_sel.first_selected_option.get_attribute('value') == '1'
        assert selenium.find_element_by_id(
            'sched_frm_active').is_selected()

    def test_13_sched_trans_modal_edit(self, base_url, selenium):
        self.get(
            selenium,
            base_url + '/payperiod/' +
            PAY_PERIOD_START_DATE.strftime('%Y-%m-%d')
        )
        link = selenium.find_element_by_xpath(
            '//a[text()="ST7 per_period (7)"]')
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        desc = body.find_element_by_id('sched_frm_description')
        desc.send_keys('edited')
        # submit the form
        selenium.find_element_by_id('modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements_by_tag_name('div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved ScheduledTransaction 7 ' \
                                 'in database.'
        # dismiss the modal
        selenium.find_element_by_id('modalCloseButton').click()
        self.wait_for_load_complete(selenium)
        # test that updated budget was removed from the page
        table = selenium.find_element_by_id('trans-table')
        texts = self.tbody2textlist(table)
        # sort order changes when we make this active
        assert texts[0][2] == '(sched) ST7 per_periodedited (7)'

    def test_14_sched_trans_modal_verify_db(self, testdb):
        t = testdb.query(ScheduledTransaction).get(7)
        assert t is not None
        assert t.description == 'ST7 per_periodedited'
        assert t.num_per_period == 2
        assert t.date is None
        assert t.day_of_month is None
        assert float(t.amount) == 11.11
        assert t.account_id == 1
        assert t.budget_id == 1
        assert t.notes == ''
        assert t.is_active is True


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb')
class TestMakeTransModal(AcceptanceHelper):

    def test_00_inactivate_scheduled(self, testdb):
        for s in testdb.query(
                ScheduledTransaction).filter(
            ScheduledTransaction.is_active.__eq__(True)
        ).all():
            s.is_active = False
            testdb.add(s)
        testdb.flush()
        testdb.commit()
        # delete existing transactions
        for tr in testdb.query(TxnReconcile).all():
            testdb.delete(tr)
        for idx in [1, 2, 3]:
            t = testdb.query(Transaction).get(idx)
            testdb.delete(t)
        testdb.flush()
        testdb.commit()

    def test_01_add_sched_transaction(self, testdb):
        pp = BiweeklyPayPeriod(PAY_PERIOD_START_DATE, testdb)
        ppdate = pp.start_date
        acct = testdb.query(Account).get(1)
        e1budget = testdb.query(Budget).get(1)
        testdb.add(ScheduledTransaction(
            account=acct,
            budget=e1budget,
            amount=11.11,
            num_per_period=2,
            description='ST7 per_period'
        ))
        testdb.flush()
        testdb.commit()
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=6)),
            actual_amount=111.13,
            budgeted_amount=111.11,
            description='T1foo',
            notes='notesT1',
            account=acct,
            scheduled_trans=testdb.query(ScheduledTransaction).get(1),
            budget=e1budget
        ))
        testdb.add(Transaction(
            date=(ppdate + timedelta(days=2)),
            actual_amount=-333.33,
            budgeted_amount=-333.33,
            description='T2',
            notes='notesT2',
            account=testdb.query(Account).get(2),
            scheduled_trans=testdb.query(ScheduledTransaction).get(3),
            budget=testdb.query(Budget).get(4)
        ))
        testdb.add(Transaction(
            date=ppdate,
            actual_amount=222.22,
            description='T3',
            notes='notesT3',
            account=testdb.query(Account).get(3),
            budget=testdb.query(Budget).get(2)
        ))

    def test_02_transaction_table(self, base_url, selenium, testdb):
        pp = BiweeklyPayPeriod(PAY_PERIOD_START_DATE, testdb)
        self.get(
            selenium,
            base_url + '/payperiod/' +
            PAY_PERIOD_START_DATE.strftime('%Y-%m-%d')
        )
        table = selenium.find_element_by_id('trans-table')
        elems = self.tbody2elemlist(table)
        htmls = []
        for row in elems:
            htmls.append(
                [x.get_attribute('innerHTML') for x in row]
            )
        assert htmls == [
            [
                '',
                '$11.11',
                '<em>(sched)</em> <a href="javascript:schedModal(7, null);">'
                'ST7 per_period (7)</a>',
                '<a href="/accounts/1">BankOne</a>',
                '<a href="/budgets/1">Periodic1</a>',
                '<a href="javascript:schedToTransModal(7, \'%s\');">make '
                'trans.</a>' % pp.start_date.strftime('%Y-%m-%d'),
                '&nbsp;'
            ],
            [
                '',
                '$11.11',
                '<em>(sched)</em> <a href="javascript:schedModal(7, null);">'
                'ST7 per_period (7)</a>',
                '<a href="/accounts/1">BankOne</a>',
                '<a href="/budgets/1">Periodic1</a>',
                '<a href="javascript:schedToTransModal(7, \'%s\');">make '
                'trans.</a>' % pp.start_date.strftime('%Y-%m-%d'),
                '&nbsp;'
            ]
        ]

    def test_03_verify_db(self, testdb):
        t = testdb.query(ScheduledTransaction).get(7)
        assert t is not None
        assert t.description == 'ST7 per_period'
        assert t.num_per_period == 2
        assert t.date is None
        assert t.day_of_month is None
        assert float(t.amount) == 11.11
        assert t.account_id == 1
        assert t.budget_id == 1
        assert t.notes is None
        assert t.is_active is True

    def test_04_make_trans_modal_num_per(self, base_url, selenium, testdb):
        pp = BiweeklyPayPeriod(PAY_PERIOD_START_DATE, testdb)
        self.get(
            selenium,
            base_url + '/payperiod/' +
            PAY_PERIOD_START_DATE.strftime('%Y-%m-%d')
        )
        link = selenium.find_elements_by_xpath(
            '//a[@href="javascript:schedToTransModal(7, \'%s\');"]'
            '' % pp.start_date.strftime('%Y-%m-%d'))[0]
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Scheduled Transaction 7 to Transaction'
        assert body.find_element_by_id(
            'schedtotrans_frm_id').get_attribute('value') == '7'
        assert body.find_element_by_id(
            'schedtotrans_frm_pp_date').get_attribute(
            'value') == pp.start_date.strftime('%Y-%m-%d')
        # if a num_per_period scheduled transaction, default to today
        # this default is populated by JS, which will use localtime
        assert body.find_element_by_id(
            'schedtotrans_frm_date').get_attribute(
            'value') == datetime.now().strftime('%Y-%m-%d')
        assert body.find_element_by_id(
            'schedtotrans_frm_amount').get_attribute('value') == '11.11'
        assert body.find_element_by_id(
            'schedtotrans_frm_description').get_attribute(
            'value') == 'ST7 per_period'
        assert body.find_element_by_id(
            'schedtotrans_frm_notes').get_attribute(
            'value') == ''

    def test_05_edit_modal(self, base_url, selenium, testdb):
        pp = BiweeklyPayPeriod(PAY_PERIOD_START_DATE, testdb)
        self.get(
            selenium,
            base_url + '/payperiod/' +
            PAY_PERIOD_START_DATE.strftime('%Y-%m-%d')
        )
        link = selenium.find_elements_by_xpath(
            '//a[@href="javascript:schedToTransModal(7, \'%s\');"]'
            '' % pp.start_date.strftime('%Y-%m-%d'))[0]
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        body.find_element_by_id('schedtotrans_frm_date').clear()
        body.find_element_by_id('schedtotrans_frm_date').send_keys(
            (pp.start_date + timedelta(days=2)).strftime('%Y-%m-%d')
        )
        body.find_element_by_id('schedtotrans_frm_amount').clear()
        body.find_element_by_id('schedtotrans_frm_amount').send_keys('22.22')
        body.find_element_by_id('schedtotrans_frm_description').send_keys(
            ' Trans'
        )
        body.find_element_by_id('schedtotrans_frm_notes').clear()
        body.find_element_by_id('schedtotrans_frm_notes').send_keys('T4notes')
        # submit the form
        selenium.find_element_by_id('modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements_by_tag_name('div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully created Transaction 6 for ' \
                                 'ScheduledTransaction 7.'
        # dismiss the modal
        selenium.find_element_by_id('modalCloseButton').click()
        self.wait_for_load_complete(selenium)
        # test that updated budget was removed from the page
        table = selenium.find_element_by_id('trans-table')
        texts = self.tbody2textlist(table)
        # sort order changes when we make this active
        assert texts[0][2] == '(sched) ST7 per_period (7)'
        assert texts[1][2] == 'ST7 per_period Trans (6)'

    def test_06_verify_db(self, testdb):
        t = testdb.query(ScheduledTransaction).get(7)
        assert t is not None
        assert t.description == 'ST7 per_period'
        assert t.num_per_period == 2
        assert t.date is None
        assert t.day_of_month is None
        assert float(t.amount) == 11.11
        assert t.account_id == 1
        assert t.budget_id == 1
        assert t.notes is None
        assert t.is_active is True
        pp = BiweeklyPayPeriod(PAY_PERIOD_START_DATE, testdb)
        t = testdb.query(Transaction).get(6)
        assert t is not None
        assert t.description == 'ST7 per_period Trans'
        assert t.date == (pp.start_date + timedelta(days=2))
        assert float(t.actual_amount) == 22.22
        assert float(t.budgeted_amount) == 11.11
        assert t.account_id == 1
        assert t.budget_id == 1
        assert t.scheduled_trans_id == 7
        assert t.notes == 'T4notes'



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016 Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import pytest
from biweeklybudget.tests.acceptance_helpers import AcceptanceHelper


@pytest.mark.acceptance
class TestIndexNavigation(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium, testflask, refreshdb):  # noqa
        self.baseurl = base_url
        self.get(selenium, base_url)

    def test_heading(self, selenium):
        heading = selenium.find_element_by_class_name('navbar-brand')
        assert heading.text == 'BiweeklyBudget'

    def test_nav_menu(self, selenium):
        ul = selenium.find_element_by_id('side-menu')
        assert ul is not None
        assert 'nav' in ul.get_attribute('class')
        assert ul.tag_name == 'ul'

    def test_notifications(self, selenium):
        div = selenium.find_element_by_id('notifications-row')
        assert div is not None
        assert div.get_attribute('class') == 'row'


@pytest.mark.acceptance
class TestIndexAccounts(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium, testflask, refreshdb):  # noqa
        self.baseurl = base_url
        self.get(selenium, base_url)

    def test_bank_table(self, selenium):
        table = selenium.find_element_by_xpath(
            "//div[@id='panel-bank-accounts']//table"
        )
        assert self.thead2list(table) == [
            'Account', 'Balance', 'Unreconciled', 'Difference'
        ]
        assert self.tbody2textlist(table) == [
            ['BankOne', '$12,789.01 (14 hours ago)', '$0.00', '$12,789.01'],
            ['BankTwoStale', '$100.23 (18 days ago)', '-$333.33', '$433.56']
        ]
        links = []
        tbody = table.find_element_by_tag_name('tbody')
        for tr in tbody.find_elements_by_tag_name('tr'):
            td = tr.find_elements_by_tag_name('td')[0]
            links.append(td.get_attribute('innerHTML'))
        assert links == [
            '<a href="/accounts/1">BankOne</a>',
            '<a href="/accounts/2">BankTwoStale</a>',
        ]

    def test_bank_stale_span(self, selenium):
        tbody = selenium.find_element_by_xpath(
            "//div[@id='panel-bank-accounts']//table/tbody"
        )
        rows = tbody.find_elements_by_tag_name('tr')
        bankTwoStale_bal_td = rows[1].find_elements_by_tag_name('td')[1]
        bal_span = bankTwoStale_bal_td.find_elements_by_tag_name('span')[1]
        assert bal_span.text == '(18 days ago)'
        assert bal_span.get_attribute('class') == 'data_age text-danger'

    def test_credit_table(self, selenium):
        table = selenium.find_element_by_xpath(
            "//div[@id='panel-credit-cards']//table"
        )
        assert self.thead2list(table) == [
            'Account', 'Balance', 'Available', 'Avail - Unrec'
        ]
        assert self.tbody2textlist(table) == [
            ['CreditOne', '-$952.06 (13 hours ago)', '$1,047.94', '$825.72'],
            ['CreditTwo', '-$5,498.65 (a day ago)', '$1.35', '$1.35']
        ]
        links = []
        tbody = table.find_element_by_tag_name('tbody')
        for tr in tbody.find_elements_by_tag_name('tr'):
            td = tr.find_elements_by_tag_name('td')[0]
            links.append(td.get_attribute('innerHTML'))
        assert links == [
            '<a href="/accounts/3">CreditOne</a>',
            '<a href="/accounts/4">CreditTwo</a>',
        ]

    def test_investment_table(self, selenium):
        table = selenium.find_element_by_xpath(
            "//div[@id='panel-investment']//table"
        )
        assert self.thead2list(table) == ['Account', 'Value']
        assert self.tbody2textlist(table) == [
            ['InvestmentOne', '$10,362.91 (13 days ago)']
        ]
        links = []
        tbody = table.find_element_by_tag_name('tbody')
        for tr in tbody.find_elements_by_tag_name('tr'):
            td = tr.find_elements_by_tag_name('td')[0]
            links.append(td.get_attribute('innerHTML'))
        assert links == [
            '<a href="/accounts/5">InvestmentOne</a>'
        ]


@pytest.mark.acceptance
@pytest.mark.usefixtures('refreshdb', 'testflask')
class TestIndexBudgets(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium):  # noqa
        self.baseurl = base_url
        self.get(selenium, base_url + '/')

    def test_budgets_table(self, selenium):
        stable = selenium.find_element_by_id('table-standing-budgets')
        stexts = self.tbody2textlist(stable)
        assert stexts == [
            ['Standing1 (4)', '$1,284.23'],
            ['Standing2 (5)', '$9,482.29']
        ]
        selems = self.tbody2elemlist(stable)
        assert selems[0][0].get_attribute(
            'innerHTML') == '<a href="/budgets/4">' \
                            'Standing1 (4)</a>'



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016 Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import pytest

from biweeklybudget.tests.acceptance_helpers import AcceptanceHelper
from biweeklybudget.models.budget_model import Budget


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
class TestBudgets(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium):  # noqa
        self.baseurl = base_url
        self.get(selenium, base_url + '/budgets')

    def test_heading(self, selenium):
        heading = selenium.find_element_by_class_name('navbar-brand')
        assert heading.text == 'Budgets - BiweeklyBudget'

    def test_nav_menu(self, selenium):
        ul = selenium.find_element_by_id('side-menu')
        assert ul is not None
        assert 'nav' in ul.get_attribute('class')
        assert ul.tag_name == 'ul'

    def test_notifications(self, selenium):
        div = selenium.find_element_by_id('notifications-row')
        assert div is not None
        assert div.get_attribute('class') == 'row'

    def test_initial_data(self, selenium):
        ptable = selenium.find_element_by_id('table-periodic-budgets')
        ptexts = self.tbody2textlist(ptable)
        stable = selenium.find_element_by_id('table-standing-budgets')
        stexts = self.tbody2textlist(stable)
        assert ptexts == [
            ['yes', 'Income (7) (income)', '$2,345.67'],
            ['yes', 'Periodic1 (1)', '$100.00'],
            ['yes', 'Periodic2 (2)', '$234.00'],
            ['NO', 'Periodic3 Inactive (3)', '$10.23']
        ]
        assert stexts == [
            ['yes', 'Standing1 (4)', '$1,284.23'],
            ['yes', 'Standing2 (5)', '$9,482.29'],
            ['NO', 'Standing3 Inactive (6)', '-$92.29']
        ]
        pelems = self.tbody2elemlist(ptable)
        selems = self.tbody2elemlist(stable)
        assert pelems[2][1].get_attribute(
            'innerHTML') == '<a href="javascript:budgetModal(2, null)">' \
                            'Periodic2 (2)</a>'
        assert pelems[0][1].get_attribute(
            'innerHTML') == '<a href="javascript:budgetModal(7, null)">' \
                            'Income (7)</a> <em class="text-success">' \
                            '(income)</em>'
        assert selems[0][1].get_attribute(
            'innerHTML') == '<a href="javascript:budgetModal(4, null)">' \
                            'Standing1 (4)</a>'


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
class TestEditPeriodic1(AcceptanceHelper):

    def test_0_verify_db(self, testdb):
        b = testdb.query(Budget).get(1)
        assert b is not None
        assert b.name == 'Periodic1'
        assert b.is_periodic is True
        assert b.description == 'P1desc'
        assert b.starting_balance == 100.00
        assert b.is_active is True
        assert b.is_income is False

    def test_1_populate_modal(self, base_url, selenium):
        self.get(selenium, base_url + '/budgets')
        link = selenium.find_element_by_xpath('//a[text()="Periodic1 (1)"]')
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Budget 1'
        assert selenium.find_element_by_id('budget_frm_name').get_attribute(
            'value') == 'Periodic1'
        assert selenium.find_element_by_id(
            'budget_frm_type_periodic').is_selected()
        assert selenium.find_element_by_id(
            'budget_frm_type_standing').is_selected() is False
        assert selenium.find_element_by_id(
            'budget_frm_description').get_attribute('value') == 'P1desc'
        assert selenium.find_element_by_id(
            'budget_frm_starting_balance').get_attribute('value') == '100'
        assert selenium.find_element_by_id(
            'budget_frm_group_starting_balance').is_displayed()
        assert selenium.find_element_by_id(
            'budget_frm_current_balance').get_attribute('value') == ''
        assert selenium.find_element_by_id(
            'budget_frm_group_current_balance').is_displayed() is False
        assert selenium.find_element_by_id('budget_frm_active').is_selected()
        assert selenium.find_element_by_id(
            'budget_frm_income').is_selected() is False

    def test_2_update_modal(self, base_url, selenium):
        # Fill in the form
        self.get(selenium, base_url + '/budgets')
        link = selenium.find_element_by_xpath('//a[text()="Periodic1 (1)"]')
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        name = selenium.find_element_by_id('budget_frm_name')
        name.clear()
        name.send_keys('EditedPeriodic1')
        desc = selenium.find_element_by_id('budget_frm_description')
        desc.clear()
        desc.send_keys('EditedP1desc')
        sb = selenium.find_element_by_id('budget_frm_starting_balance')
        assert sb.is_displayed()
        sb.clear()
        sb.send_keys('2345.67')
        selenium.find_element_by_id('budget_frm_active').click()
        assert selenium.find_element_by_id(
            'budget_frm_active').is_selected() is False
        assert selenium.find_element_by_id(
            'budget_frm_income').is_selected() is False
        # submit the form
        selenium.find_element_by_id('modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements_by_tag_name('div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Budget 1 in database.'
        # dismiss the modal
        selenium.find_element_by_id('modalCloseButton').click()
        self.wait_for_load_complete(selenium)
        self.wait_for_id(selenium, 'table-periodic-budgets')
        # test that updated budget was removed from the page
        ptable = selenium.find_element_by_id('table-periodic-budgets')
        pelems = self.tbody2elemlist(ptable)
        assert pelems[0][1].get_attribute(
            'innerHTML') == '<a href="javascript:budgetModal(1, null)">' \
                            'EditedPeriodic1 (1)</a>'

    def test_3_verify_db(self, testdb):
        b = testdb.query(Budget).get(1)
        assert b is not None
        assert b.name == 'EditedPeriodic1'
        assert b.is_periodic is True
        assert b.description == 'EditedP1desc'
        assert float(b.starting_balance) == 2345.6700
        assert b.is_active is False
        assert b.is_income is False


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
class TestEditPeriodic2(AcceptanceHelper):

    def test_1_populate_modal(self, base_url, selenium):
        self.get(selenium, base_url + '/budgets')
        link = selenium.find_element_by_xpath('//a[text()="Periodic2 (2)"]')
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Budget 2'
        assert selenium.find_element_by_id('budget_frm_name').get_attribute(
            'value') == 'Periodic2'
        assert selenium.find_element_by_id(
            'budget_frm_type_periodic').is_selected()
        assert selenium.find_element_by_id(
            'budget_frm_type_standing').is_selected() is False
        assert selenium.find_element_by_id(
            'budget_frm_description').get_attribute('value') == 'P2desc'
        assert selenium.find_element_by_id(
            'budget_frm_starting_balance').get_attribute('value') == '234'
        assert selenium.find_element_by_id(
            'budget_frm_group_starting_balance').is_displayed()
        assert selenium.find_element_by_id(
            'budget_frm_current_balance').get_attribute('value') == ''
        assert selenium.find_element_by_id(
            'budget_frm_group_current_balance').is_displayed() is False
        assert selenium.find_element_by_id('budget_frm_active').is_selected()
        assert selenium.find_element_by_id(
            'budget_frm_income').is_selected() is False


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
class TestEditPeriodic3(AcceptanceHelper):

    def test_1_populate_modal(self, base_url, selenium):
        self.get(selenium, base_url + '/budgets')
        link = selenium.find_element_by_xpath(
            '//a[text()="Periodic3 Inactive (3)"]'
        )
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Budget 3'
        assert selenium.find_element_by_id('budget_frm_name').get_attribute(
            'value') == 'Periodic3 Inactive'
        assert selenium.find_element_by_id(
            'budget_frm_type_periodic').is_selected()
        assert selenium.find_element_by_id(
            'budget_frm_type_standing').is_selected() is False
        assert selenium.find_element_by_id(
            'budget_frm_description').get_attribute('value') == 'P3desc'
        assert selenium.find_element_by_id(
            'budget_frm_starting_balance').get_attribute('value') == '10.23'
        assert selenium.find_element_by_id(
            'budget_frm_group_starting_balance').is_displayed()
        assert selenium.find_element_by_id(
            'budget_frm_current_balance').get_attribute('value') == ''
        assert selenium.find_element_by_id(
            'budget_frm_group_current_balance').is_displayed() is False
        assert selenium.find_element_by_id(
            'budget_frm_active').is_selected() is False
        assert selenium.find_element_by_id(
            'budget_frm_income').is_selected() is False


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
class TestEditStanding1(AcceptanceHelper):

    def test_1_populate_modal(self, base_url, selenium):
        self.get(selenium, base_url + '/budgets')
        link = selenium.find_element_by_xpath('//a[text()="Standing1 (4)"]')
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Budget 4'
        assert selenium.find_element_by_id('budget_frm_name').get_attribute(
            'value') == 'Standing1'
        assert selenium.find_element_by_id(
            'budget_frm_type_periodic').is_selected() is False
        assert selenium.find_element_by_id(
            'budget_frm_type_standing').is_selected()
        assert selenium.find_element_by_id(
            'budget_frm_description').get_attribute('value') == 'S1desc'
        assert selenium.find_element_by_id(
            'budget_frm_starting_balance').get_attribute('value') == ''
        assert selenium.find_element_by_id(
            'budget_frm_group_starting_balance').is_displayed() is False
        assert selenium.find_element_by_id(
            'budget_frm_current_balance').get_attribute('value') == '1284.23'
        assert selenium.find_element_by_id(
            'budget_frm_group_current_balance').is_displayed()
        assert selenium.find_element_by_id('budget_frm_active').is_selected()
        assert selenium.find_element_by_id(
            'budget_frm_income').is_selected() is False


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
class TestEditStanding2(AcceptanceHelper):

    def test_1_populate_modal(self, base_url, selenium):
        self.get(selenium, base_url + '/budgets')
        link = selenium.find_element_by_xpath('//a[text()="Standing2 (5)"]')
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Budget 5'
        assert selenium.find_element_by_id('budget_frm_name').get_attribute(
            'value') == 'Standing2'
        assert selenium.find_element_by_id(
            'budget_frm_type_periodic').is_selected() is False
        assert selenium.find_element_by_id(
            'budget_frm_type_standing').is_selected()
        assert selenium.find_element_by_id(
            'budget_frm_description').get_attribute('value') == 'S2desc'
        assert selenium.find_element_by_id(
            'budget_frm_starting_balance').get_attribute('value') == ''
        assert selenium.find_element_by_id(
            'budget_frm_group_starting_balance').is_displayed() is False
        assert selenium.find_element_by_id(
            'budget_frm_current_balance').get_attribute('value') == '9482.29'
        assert selenium.find_element_by_id(
            'budget_frm_group_current_balance').is_displayed()
        assert selenium.find_element_by_id('budget_frm_active').is_selected()
        assert selenium.find_element_by_id(
            'budget_frm_income').is_selected() is False


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
class TestEditStanding3(AcceptanceHelper):

    def test_1_populate_modal(self, base_url, selenium):
        self.get(selenium, base_url + '/budgets')
        link = selenium.find_element_by_xpath(
            '//a[text()="Standing3 Inactive (6)"]'
        )
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Budget 6'
        assert selenium.find_element_by_id('budget_frm_name').get_attribute(
            'value') == 'Standing3 Inactive'
        assert selenium.find_element_by_id(
            'budget_frm_type_periodic').is_selected() is False
        assert selenium.find_element_by_id(
            'budget_frm_type_standing').is_selected()
        assert selenium.find_element_by_id(
            'budget_frm_description').get_attribute('value') == 'S3desc'
        assert selenium.find_element_by_id(
            'budget_frm_starting_balance').get_attribute('value') == ''
        assert selenium.find_element_by_id(
            'budget_frm_group_starting_balance').is_displayed() is False
        assert selenium.find_element_by_id(
            'budget_frm_current_balance').get_attribute('value') == '-92.29'
        assert selenium.find_element_by_id(
            'budget_frm_group_current_balance').is_displayed()
        assert selenium.find_element_by_id(
            'budget_frm_active').is_selected() is False
        assert selenium.find_element_by_id(
            'budget_frm_income').is_selected() is False


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
class TestEditIncome(AcceptanceHelper):

    def test_0_verify_db(self, testdb):
        b = testdb.query(Budget).get(7)
        assert b is not None
        assert b.name == 'Income'
        assert b.is_periodic is True
        assert b.description == 'IncomeDesc'
        assert float(b.starting_balance) == 2345.67
        assert b.is_active is True
        assert b.is_income is True

    def test_1_populate_modal(self, base_url, selenium):
        self.get(selenium, base_url + '/budgets')
        link = selenium.find_element_by_xpath('//a[text()="Income (7)"]')
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Budget 7'
        assert selenium.find_element_by_id('budget_frm_name').get_attribute(
            'value') == 'Income'
        assert selenium.find_element_by_id(
            'budget_frm_type_periodic').is_selected()
        assert selenium.find_element_by_id(
            'budget_frm_type_standing').is_selected() is False
        assert selenium.find_element_by_id(
            'budget_frm_description').get_attribute('value') == 'IncomeDesc'
        assert selenium.find_element_by_id(
            'budget_frm_starting_balance').get_attribute('value') == '2345.67'
        assert selenium.find_element_by_id(
            'budget_frm_group_starting_balance').is_displayed()
        assert selenium.find_element_by_id(
            'budget_frm_current_balance').get_attribute('value') == ''
        assert selenium.find_element_by_id(
            'budget_frm_group_current_balance').is_displayed() is False
        assert selenium.find_element_by_id('budget_frm_active').is_selected()
        assert selenium.find_element_by_id('budget_frm_income').is_selected()

    def test_2_update_modal(self, base_url, selenium):
        # Fill in the form
        self.get(selenium, base_url + '/budgets')
        link = selenium.find_element_by_xpath('//a[text()="Income (7)"]')
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        name = selenium.find_element_by_id('budget_frm_name')
        name.clear()
        name.send_keys('EditedIncome')
        desc = selenium.find_element_by_id('budget_frm_description')
        desc.send_keys('edited')
        sb = selenium.find_element_by_id('budget_frm_starting_balance')
        assert sb.is_displayed()
        sb.clear()
        sb.send_keys('123.45')
        selenium.find_element_by_id('budget_frm_active').click()
        assert selenium.find_element_by_id(
            'budget_frm_active').is_selected() is False
        assert selenium.find_element_by_id(
            'budget_frm_income').is_selected() is True
        # submit the form
        selenium.find_element_by_id('modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements_by_tag_name('div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Budget 7 in database.'
        # dismiss the modal
        selenium.find_element_by_id('modalCloseButton').click()
        self.wait_for_load_complete(selenium)
        self.wait_for_id(selenium, 'table-periodic-budgets')
        # test that updated budget was removed from the page
        ptable = selenium.find_element_by_id('table-periodic-budgets')
        pelems = self.tbody2elemlist(ptable)
        assert pelems[0][1].get_attribute(
            'innerHTML') == '<a href="javascript:budgetModal(7, null)">' \
                            'EditedIncome (7)</a> ' \
                            '<em class="text-success">(income)</em>'

    def test_3_verify_db(self, testdb):
        b = testdb.query(Budget).get(7)
        assert b is not None
        assert b.name == 'EditedIncome'
        assert b.is_periodic is True
        assert b.description == 'IncomeDescedited'
        assert float(b.starting_balance) == 123.45
        assert b.is_active is False
        assert b.is_income is True


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
class TestDirectURLPeriodic1(AcceptanceHelper):

    def test_1_populate_modal(self, base_url, selenium):
        self.get(selenium, base_url + '/budgets/1')
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Edit Budget 1'
        assert selenium.find_element_by_id('budget_frm_name').get_attribute(
            'value') == 'Periodic1'
        assert selenium.find_element_by_id(
            'budget_frm_type_periodic').is_selected()
        assert selenium.find_element_by_id(
            'budget_frm_type_standing').is_selected() is False
        assert selenium.find_element_by_id(
            'budget_frm_description').get_attribute('value') == 'P1desc'
        assert selenium.find_element_by_id(
            'budget_frm_starting_balance').get_attribute('value') == '100'
        assert selenium.find_element_by_id(
            'budget_frm_group_starting_balance').is_displayed()
        assert selenium.find_element_by_id(
            'budget_frm_current_balance').get_attribute('value') == ''
        assert selenium.find_element_by_id(
            'budget_frm_group_current_balance').is_displayed() is False
        assert selenium.find_element_by_id('budget_frm_active').is_selected()
        assert selenium.find_element_by_id(
            'budget_frm_income').is_selected() is False


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
class TestAddStandingBudget(AcceptanceHelper):

    def test_2_update_modal(self, base_url, selenium):
        # Fill in the form
        self.get(selenium, base_url + '/budgets')
        link = selenium.find_element_by_id('btn_add_budget')
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        name = selenium.find_element_by_id('budget_frm_name')
        name.clear()
        name.send_keys('NewStanding')
        standing = selenium.find_element_by_id('budget_frm_type_standing')
        standing.click()
        desc = selenium.find_element_by_id('budget_frm_description')
        desc.clear()
        desc.send_keys('Newly Added Standing')
        sb = selenium.find_element_by_id('budget_frm_current_balance')
        assert sb.is_displayed()
        sb.clear()
        sb.send_keys('6789.12')
        assert selenium.find_element_by_id(
            'budget_frm_active').is_selected()
        assert selenium.find_element_by_id(
            'budget_frm_income').is_selected() is False
        # submit the form
        selenium.find_element_by_id('modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements_by_tag_name('div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Budget 8 in database.'
        # dismiss the modal
        selenium.find_element_by_id('modalCloseButton').click()
        self.wait_for_load_complete(selenium)
        self.wait_for_id(selenium, 'table-periodic-budgets')
        # test that updated budget was removed from the page
        stable = selenium.find_element_by_id('table-standing-budgets')
        selems = self.tbody2elemlist(stable)
        assert selems[0][1].get_attribute(
            'innerHTML') == '<a href="javascript:budgetModal(8, null)">' \
                            'NewStanding (8)</a>'

    def test_3_verify_db(self, testdb):
        b = testdb.query(Budget).get(8)
        assert b is not None
        assert b.name == 'NewStanding'
        assert b.is_periodic is False
        assert b.description == 'Newly Added Standing'
        assert float(b.current_balance) == 6789.12
        assert b.is_active is True
        assert b.is_income is False


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
class TestAddIncomeBudget(AcceptanceHelper):

    def test_2_update_modal(self, base_url, selenium):
        # Fill in the form
        self.get(selenium, base_url + '/budgets')
        link = selenium.find_element_by_id('btn_add_budget')
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        name = selenium.find_element_by_id('budget_frm_name')
        name.clear()
        name.send_keys('NewIncome')
        periodic = selenium.find_element_by_id('budget_frm_type_periodic')
        periodic.click()
        desc = selenium.find_element_by_id('budget_frm_description')
        desc.clear()
        desc.send_keys('Newly Added Income')
        sb = selenium.find_element_by_id('budget_frm_starting_balance')
        assert sb.is_displayed()
        sb.clear()
        sb.send_keys('123.45')
        assert selenium.find_element_by_id(
            'budget_frm_active').is_selected()
        income = selenium.find_element_by_id('budget_frm_income')
        income.click()
        assert income.is_selected()
        # submit the form
        selenium.find_element_by_id('modalSaveButton').click()
        self.wait_for_jquery_done(selenium)
        # check that we got positive confirmation
        _, _, body = self.get_modal_parts(selenium)
        x = body.find_elements_by_tag_name('div')[0]
        assert 'alert-success' in x.get_attribute('class')
        assert x.text.strip() == 'Successfully saved Budget 8 in database.'
        # dismiss the modal
        selenium.find_element_by_id('modalCloseButton').click()
        self.wait_for_load_complete(selenium)
        self.wait_for_id(selenium, 'table-periodic-budgets')
        # test that updated budget was removed from the page
        stable = selenium.find_element_by_id('table-periodic-budgets')
        selems = self.tbody2elemlist(stable)
        assert selems[1][1].get_attribute(
            'innerHTML') == '<a href="javascript:budgetModal(8, null)">' \
                            'NewIncome (8)</a> <em class="text-success">' \
                            '(income)</em>'

    def test_3_verify_db(self, testdb):
        b = testdb.query(Budget).get(8)
        assert b is not None
        assert b.name == 'NewIncome'
        assert b.is_periodic is True
        assert b.description == 'Newly Added Income'
        assert float(b.starting_balance) == 123.45
        assert b.is_active is True
        assert b.is_income is True


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
class TestBudgetOverBalanceNotification(AcceptanceHelper):

    def test_0_update_db(self, testdb):
        b = testdb.query(Budget).get(4)
        b.current_balance = 123456.78
        testdb.add(b)
        testdb.flush()
        testdb.commit()

    def test_1_notification(self, base_url, selenium):
        self.baseurl = base_url
        # Fill in the form
        self.get(selenium, base_url + '/budgets')
        stable = selenium.find_element_by_id('table-standing-budgets')
        stexts = self.tbody2textlist(stable)
        assert stexts[0] == ['yes', 'Standing1 (4)', '$123,456.78']
        div = selenium.find_elements_by_xpath(
            "//div[@id='notifications-row']/div/div"
        )[1]
        assert div.text == 'Combined balance of all budget-funding accounts ' \
                           '($12,889.24) is less than balance of all ' \
                           'standing budgets ($132,939.07)!'
        a = div.find_elements_by_tag_name('a')
        assert self.relurl(a[0].get_attribute('href')) == '/accounts'
        assert a[0].text == 'budget-funding accounts'
        assert self.relurl(a[1].get_attribute('href')) == '/budgets'
        assert a[1].text == 'standing budgets'



"""
The latest version of this package is available at:
<http://github.com/jantman/biweeklybudget>

################################################################################
Copyright 2016 Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>

    This file is part of biweeklybudget, also known as biweeklybudget.

    biweeklybudget is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    biweeklybudget is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with biweeklybudget.  If not, see <http://www.gnu.org/licenses/>.

The Copyright and Authors attributions contained herein may not be removed or
otherwise altered, except to add the Author attribution of a contributor to
this work. (Additional Terms pursuant to Section 7b of the AGPL v3)
################################################################################
While not legally required, I sincerely request that anyone who finds
bugs please submit them at <https://github.com/jantman/biweeklybudget> or
to me via email, and that you send any contributions or improvements
either as a pull request on GitHub, or to me via email.
################################################################################

AUTHORS:
Jason Antman <jason@jasonantman.com> <http://www.jasonantman.com>
################################################################################
"""

import pytest
from datetime import timedelta, datetime
from selenium.webdriver.support.ui import Select
from pytz import UTC

from biweeklybudget.utils import dtnow
from biweeklybudget.models.txn_reconcile import TxnReconcile
from biweeklybudget.tests.acceptance_helpers import AcceptanceHelper


@pytest.mark.acceptance
class TestOFX(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium, testflask, refreshdb):  # noqa
        self.baseurl = base_url
        self.get(selenium, base_url + '/ofx')

    def test_heading(self, selenium):
        heading = selenium.find_element_by_class_name('navbar-brand')
        assert heading.text == 'OFX Transactions - BiweeklyBudget'

    def test_nav_menu(self, selenium):
        ul = selenium.find_element_by_id('side-menu')
        assert ul is not None
        assert 'nav' in ul.get_attribute('class')
        assert ul.tag_name == 'ul'

    def test_notifications(self, selenium):
        div = selenium.find_element_by_id('notifications-row')
        assert div is not None
        assert div.get_attribute('class') == 'row'


@pytest.mark.acceptance
class TestOFXDefault(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium, testflask, refreshdb):  # noqa
        self.baseurl = base_url
        self.dt = dtnow()
        self.get(selenium, base_url + '/ofx')

    def test_table(self, selenium):
        table = selenium.find_element_by_id('table-ofx-txn')
        texts = self.tbody2textlist(table)
        elems = self.tbody2elemlist(table)
        assert texts[0] == [
            (self.dt - timedelta(hours=22)).strftime('%Y-%m-%d'),
            '$123.81',
            'CreditOne (3)',
            'Purchase',
            '123.81 Credit Purchase T1',
            '38328',
            'CreditOneT1Desc',
            'T1',
            '4',
            (self.dt - timedelta(hours=13)).strftime('%Y-%m-%d'),
            ''
        ]
        assert elems[0][2].get_attribute(
            'innerHTML') == '<a href="/accounts/3">CreditOne (3)</a>'
        assert elems[0][7].get_attribute(
            'innerHTML') == '<a href="javascript:ofxTransModal(' \
                            '3, \'T1\')">T1</a>'
        trans = [[t[2], t[7]] for t in texts]
        assert trans == [
            ['CreditOne (3)', 'T1'],
            ['CreditTwo (4)', '002'],
            ['BankOne (1)', 'BankOne.0.1'],
            ['CreditTwo (4)', '001'],
            ['BankOne (1)', 'BankOne.1.1'],
            ['BankOne (1)', 'BankOne.1.2'],
            ['BankOne (1)', 'BankOne.1.3'],
            ['BankOne (1)', 'BankOne.1.4'],
            ['BankOne (1)', 'BankOne.1.5'],
            ['BankOne (1)', 'BankOne.1.6']
        ]
        reconciles = [t[10] for t in texts]
        assert reconciles == [
            '',
            '',
            'Yes (1)',
            '',
            '',
            '',
            '',
            '',
            '',
            ''
        ]
        assert elems[2][10].get_attribute('innerHTML') == '<a ' \
            'href="javascript:txnReconcileModal(1)">Yes (1)</a>'

    def test_paginate(self, selenium):
        self.get(selenium, self.baseurl + '/ofx')
        table = self.retry_stale(selenium.find_element_by_id, 'table-ofx-txn')
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [[t[2], t[7]] for t in texts]
        assert trans == [
            ['CreditOne (3)', 'T1'],
            ['CreditTwo (4)', '002'],
            ['BankOne (1)', 'BankOne.0.1'],
            ['CreditTwo (4)', '001'],
            ['BankOne (1)', 'BankOne.1.1'],
            ['BankOne (1)', 'BankOne.1.2'],
            ['BankOne (1)', 'BankOne.1.3'],
            ['BankOne (1)', 'BankOne.1.4'],
            ['BankOne (1)', 'BankOne.1.5'],
            ['BankOne (1)', 'BankOne.1.6']
        ]
        page2_link = selenium.find_element_by_xpath(
            '//li[@class="paginate_button "]/a[text()="2"]'
        )
        page2_link.click()
        table = self.retry_stale(selenium.find_element_by_id, 'table-ofx-txn')
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [[t[2], t[7]] for t in texts]
        assert trans == [
            ['BankOne (1)', 'BankOne.1.7'],
            ['BankOne (1)', 'BankOne.1.8'],
            ['BankOne (1)', 'BankOne.1.9'],
            ['BankOne (1)', 'BankOne.1.10'],
            ['BankOne (1)', 'BankOne.1.11'],
            ['BankOne (1)', 'BankOne.1.12'],
            ['BankOne (1)', 'BankOne.1.13'],
            ['BankOne (1)', 'BankOne.1.14'],
            ['BankOne (1)', 'BankOne.1.15'],
            ['BankOne (1)', 'BankOne.1.16']
        ]

    def test_filter_opts(self, selenium):
        self.get(selenium, self.baseurl + '/ofx')
        acct_filter = Select(selenium.find_element_by_id('account_filter'))
        # find the options
        opts = []
        for o in acct_filter.options:
            opts.append([o.get_attribute('value'), o.text])
        assert opts == [
            ['None', ''],
            ['1', 'BankOne'],
            ['2', 'BankTwoStale'],
            ['3', 'CreditOne'],
            ['4', 'CreditTwo'],
            ['6', 'DisabledBank'],
            ['5', 'InvestmentOne']
        ]

    def test_filter(self, selenium):
        p1trans = [
            ['CreditOne (3)', 'T1'],
            ['CreditTwo (4)', '002'],
            ['BankOne (1)', 'BankOne.0.1'],
            ['CreditTwo (4)', '001'],
            ['BankOne (1)', 'BankOne.1.1'],
            ['BankOne (1)', 'BankOne.1.2'],
            ['BankOne (1)', 'BankOne.1.3'],
            ['BankOne (1)', 'BankOne.1.4'],
            ['BankOne (1)', 'BankOne.1.5'],
            ['BankOne (1)', 'BankOne.1.6']
        ]
        self.get(selenium, self.baseurl + '/ofx')
        table = self.retry_stale(selenium.find_element_by_id, 'table-ofx-txn')
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [[t[2], t[7]] for t in texts]
        # check sanity
        assert trans == p1trans
        acct_filter = Select(selenium.find_element_by_id('account_filter'))
        # select CreditTwo (4)
        acct_filter.select_by_value('4')
        table = self.retry_stale(selenium.find_element_by_id, 'table-ofx-txn')
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [[t[2], t[7]] for t in texts]
        assert trans == [
            ['CreditTwo (4)', '002'],
            ['CreditTwo (4)', '001']
        ]
        # select back to all
        acct_filter.select_by_value('None')
        table = self.retry_stale(selenium.find_element_by_id, 'table-ofx-txn')
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [[t[2], t[7]] for t in texts]
        assert trans == p1trans

    def test_search(self, selenium):
        p1trans = [
            ['CreditTwo (4)', '001'],
            ['BankTwoStale (2)', '1'],
            ['DisabledBank (6)', '001']
        ]
        self.get(selenium, self.baseurl + '/ofx')
        search = self.retry_stale(
            selenium.find_element_by_xpath,
            '//input[@type="search"]'
        )
        search.send_keys('inte')
        table = self.retry_stale(selenium.find_element_by_id, 'table-ofx-txn')
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [[t[2], t[7]] for t in texts]
        # check sanity
        assert trans == p1trans
        acct_filter = Select(selenium.find_element_by_id('account_filter'))
        # select CreditTwo (4)
        acct_filter.select_by_value('4')
        table = self.retry_stale(selenium.find_element_by_id, 'table-ofx-txn')
        texts = self.retry_stale(self.tbody2textlist, table)
        trans = [[t[2], t[7]] for t in texts]
        assert trans == [
            ['CreditTwo (4)', '001']
        ]


@pytest.mark.acceptance
class TestOFXTransModal(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium, testflask, refreshdb):  # noqa
        self.baseurl = base_url
        self.get(selenium, base_url + '/ofx')

    def test_modal_on_click(self, selenium):
        link = selenium.find_element_by_xpath('//a[text()="T1"]')
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'OFXTransaction Account=3 FITID=T1'
        texts = self.tbody2textlist(body)
        elems = self.tbody2elemlist(body)
        pdate = (dtnow() - timedelta(hours=22))
        fdate = (dtnow() - timedelta(hours=13))
        assert texts == [
            ['Account', 'CreditOne (3)'],
            ['FITID', 'T1'],
            ['Date Posted', pdate.strftime('%Y-%m-%d')],
            ['Amount', '$123.81'],
            ['Name', '123.81 Credit Purchase T1'],
            ['Memo', '38328'],
            ['Type', 'Purchase'],
            ['Description', 'CreditOneT1Desc'],
            ['Notes', ''],
            ['Checknum', '123'],
            ['MCC', 'T1mcc'],
            ['SIC', 'T1sic'],
            ['OFX Statement'],
            ['ID', '4'],
            ['Date', fdate.strftime('%Y-%m-%d')],
            ['Filename', '/stmt/CreditOne/0'],
            ['File mtime', fdate.strftime('%Y-%m-%d')],
            ['Ledger Balance', '-$952.06']
        ]
        assert elems[0][1].get_attribute(
            'innerHTML') == '<a href="/accounts/3">CreditOne (3)</a>'
        assert selenium.find_element_by_id(
            'modalSaveButton').is_displayed() is False


@pytest.mark.acceptance
class TestOFXTransURL(AcceptanceHelper):

    @pytest.fixture(autouse=True)
    def get_page(self, base_url, selenium, testflask, refreshdb):  # noqa
        self.baseurl = base_url
        self.get(selenium, base_url + '/ofx/3/T1')

    def test_modal_auto_displayed(self, selenium):
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'OFXTransaction Account=3 FITID=T1'
        texts = self.tbody2textlist(body)
        elems = self.tbody2elemlist(body)
        pdate = (dtnow() - timedelta(hours=22))
        fdate = (dtnow() - timedelta(hours=13))
        assert texts == [
            ['Account', 'CreditOne (3)'],
            ['FITID', 'T1'],
            ['Date Posted', pdate.strftime('%Y-%m-%d')],
            ['Amount', '$123.81'],
            ['Name', '123.81 Credit Purchase T1'],
            ['Memo', '38328'],
            ['Type', 'Purchase'],
            ['Description', 'CreditOneT1Desc'],
            ['Notes', ''],
            ['Checknum', '123'],
            ['MCC', 'T1mcc'],
            ['SIC', 'T1sic'],
            ['OFX Statement'],
            ['ID', '4'],
            ['Date', fdate.strftime('%Y-%m-%d')],
            ['Filename', '/stmt/CreditOne/0'],
            ['File mtime', fdate.strftime('%Y-%m-%d')],
            ['Ledger Balance', '-$952.06']
        ]
        assert elems[0][1].get_attribute(
            'innerHTML') == '<a href="/accounts/3">CreditOne (3)</a>'
        assert selenium.find_element_by_id(
            'modalSaveButton').is_displayed() is False


@pytest.mark.acceptance
@pytest.mark.usefixtures('class_refresh_db', 'refreshdb', 'testflask')
class TestTransReconciledModal(AcceptanceHelper):

    def test_0_verify_db(self, testdb):
        t = testdb.query(TxnReconcile).get(1)
        assert t.ofx_account_id == 1
        assert t.ofx_fitid == 'BankOne.0.1'
        assert t.txn_id == 1
        assert t.rule_id is None
        assert t.note == 'reconcile notes'
        assert t.reconciled_at == datetime(2017, 4, 10, 8, 9, 11, tzinfo=UTC)

    def test_1_modal(self, base_url, selenium):
        self.baseurl = base_url
        self.get(selenium, base_url + '/transactions')
        link = selenium.find_element_by_xpath(
            '//a[@href="javascript:txnReconcileModal(1)"]')
        link.click()
        modal, title, body = self.get_modal_parts(selenium)
        self.assert_modal_displayed(modal, title, body)
        assert title.text == 'Transaction Reconcile 1'
        dl = body.find_element_by_tag_name('dl')
        assert dl.get_attribute('innerHTML') == '\n' \
            '<dt>Date Reconciled</dt><dd>2017-04-10 08:09:11 UTC</dd>\n' \
            '<dt>Note</dt><dd>reconcile notes</dd>\n' \
            '<dt>Rule</dt><dd>null</dd>\n'
        trans_tbl = body.find_element_by_id('txnReconcileModal-trans')
        trans_texts = self.tbody2textlist(trans_tbl)
        assert trans_texts == [
            ['Transaction'],
            [
                'Date',
                (dtnow() + timedelta(days=4)).strftime('%Y-%m-%d')
            ],
            ['Amount', '$111.13'],
            ['Budgeted Amount', '$111.11'],
            ['Description', 'T1foo'],
            ['Account', 'BankOne (1)'],
            ['Budget', 'Periodic1 (1)'],
            ['Notes', 'notesT1'],
            ['Scheduled?', 'Yes (1)']
        ]
        trans_elems = self.tbody2elemlist(trans_tbl)
        assert trans_elems[5][1].get_attribute('innerHTML') == '<a href=' \
            '"/accounts/1">BankOne (1)</a>'
        assert trans_elems[6][1].get_attribute('innerHTML') == '<a href=' \
            '"/budgets/1">Periodic1 (1)</a>'
        assert trans_elems[8][1].get_attribute('innerHTML') == '<a href=' \
            '"/scheduled/1">Yes (1)</a>'
        ofx_tbl = body.find_element_by_id('txnReconcileModal-ofx')
        ofx_texts = self.tbody2textlist(ofx_tbl)
        assert ofx_texts == [
            ['OFX Transaction'],
            ['Account', 'BankOne (1)'],
            ['FITID', 'BankOne.0.1'],
            ['Date Posted', (dtnow() - timedelta(days=6)).strftime('%Y-%m-%d')],
            ['Amount', '-$20.00'],
            ['Name', 'Late Fee'],
            ['Memo', ''],
            ['Type', 'Debit'],
            ['Description', ''],
            ['Notes', ''],
            ['Checknum', ''],
            ['MCC', ''],
            ['SIC', ''],
            ['OFX Statement'],
            ['ID', '1'],
            ['Date', (dtnow() - timedelta(hours=46)).strftime('%Y-%m-%d')],
            ['Filename', '/stmt/BankOne/0'],
            [
                'File mtime',
                (dtnow() - timedelta(hours=46)).strftime('%Y-%m-%d')
            ],
            ['Ledger Balance', '$12,345.67']
        ]
        ofx_elems = self.tbody2elemlist(ofx_tbl)
        assert ofx_elems[1][1].get_attribute('innerHTML') == '<a href=' \
            '"/accounts/1">BankOne (1)</a>'
